!
! Module: array_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to process arrays in various simple ways. The criteria for a subroutine or function to be placed here rather than elsewhere are:
!	(i) One of the input or output arguments must be an array, of any dimension >= 1, of any data type;
!	(ii) The operation performed on the array should either be a logical or simple arithmetic one.
! 
! Comments: TODO:
! 
! $Id: array_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module array_utils

  use types		! For single, double, int16 etc type definitions.

  use geometric_types	! for Point2dT type definition.
  use ssc_misc		! for swap.
  use sort_mod		! for findUniqueValues.

  implicit none 

  private 

  ! Take array, return indices:
  public maskArrayLimits
  public largestTrueIndex
  public smallestTrueIndex
  public getBinNumber
  public maxLoc1d
  public minLoc1d

  ! Take and return array:
  public findUniqueValues
  public vectorCross
  public normalizeVector
  public findEdges
  public selectArraySubset
  public cyclicConvolution
!  public scaleBalance
  public rotateArray

  ! Take array, return scalars:
  public linearInterpolate
  public trapezoidSum
  public getGridDiffs
  public interpolateGridDiff
!  public centreHistoMassMedian
  public median
  public valueAtHistoFraction
!  public preciseSum
!** can't see that this is any better than sum().
  public angleBetween

  interface largestTrueIndex
    module procedure largestTrueIndexBool, largestTrueIndexDefault
  end interface
  interface smallestTrueIndex
    module procedure smallestTrueIndexBool, smallestTrueIndexDefault
  end interface
  ! NOTE!!! At present 'logical' and 'logical(bool)' are different. If 'types' ever defines them the same, then this overloading will have to be deleted.

  interface getBinNumber
    module procedure getBinNumberSingle, getBinNumberDouble
  end interface

  interface maxLoc1d
    module procedure maxLoc1dSingle, maxLoc1dDouble
  end interface
  interface minLoc1d
    module procedure minLoc1dSingle, minLoc1dDouble
  end interface

  interface findUniqueValues
    module procedure findUniqueValuesInt8, findUniqueValuesInt16
    module procedure findUniqueValuesInt32
  end interface

  interface vectorCross
    module procedure vectorCrossSingle, vectorCrossDouble
  end interface

  interface normalizeVector
    module procedure normalizeVectorSingle, normalizeVectorDouble
  end interface

  interface selectArraySubset
    module procedure selectArraySubsetInt8
    module procedure selectArraySubsetInt16
    module procedure selectArraySubsetInt32
    module procedure selectArraySubsetSingle
    module procedure selectArraySubsetDouble
    module procedure selectArraySubsetBool
  end interface

  interface rotateArray
    module procedure rotateArraySingle
  end interface

  interface trapezoidSum
    module procedure trapezoidSumSingle, trapezoidSumDouble
  end interface

!  interface centreHistoMassMedian
!    module procedure cohmMedianInt8Vector, cohmMedianInt8Array
!    module procedure cohmMedianInt16Vector, cohmMedianInt16Array
!    module procedure cohmMedianInt32Vector, cohmMedianInt32Array
!  end interface

  interface median
    module procedure medianSingleVector, medianSingleArray
    module procedure medianDoubleVector, medianDoubleArray
  end interface

  interface valueAtHistoFraction
    module procedure valueAtHistoFracSingleArr, valueAtHistoFracDoubleArr
    module procedure valueAtHistoFracSingleVec, valueAtHistoFracDoubleVec
  end interface

!  interface preciseSum
!    module procedure preciseSumVectorSingle, preciseSumVectorDouble
!    module procedure preciseSumArraySingle, preciseSumArrayDouble
!  end interface

  interface angleBetween
    module procedure angleBetweenSingle, angleBetweenDouble
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine maskArrayLimits(mask, minXi, maxXi, minYi, maxYi, status)
  !---------------------------------------------------------------------
  ! Adapted from CGP's im_lims subroutine in implot-1.0.
  !
  ! Given a 2D logical array mask(), this subroutine returns the following:
  !
  !	minXi = the index of the first column which contains at least 1
  !	TRUE value;
  !
  !	maxXi = the index of the last column which contains at least 1
  !	TRUE value;
  !
  !	minYi = the index of the first row which contains at least 1 TRUE
  !	value;
  !
  !	maxYi = the index of the last row which contains at least 1 TRUE
  !	value.
  !
  ! Note that the index of the first row/column is 1, also that mask() is
  ! assumed to follow the position-vector format, ie the first index of
  ! mask() is the column number and the second the row number.

    implicit none

    logical(bool),  intent(in)            :: mask(:,:)
    integer(int16), intent(out)           :: minXi, maxXi, minYi, maxYi
    integer,        intent(out), optional :: status

    integer(int16) :: localStatus, xSize, ySize, xi, yi

    !-------------------------------------------------------------------
    ! Start

    ! Defaults:
    localStatus = 0
    minXi = 1
    maxXi = 0
    minYi = 1
    maxYi = 0
    
    if (count(mask) .le. 0) then
      if (present(status)) status = -1
  return
    end if

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    if (xSize .le. 0 .or. ySize .le. 0) then
      if (present(status)) status = -2
  return
    end if

    do xi = 1, xSize
      if (any(mask(xi, :))) then
        minXi = xi
    exit
      end if
    end do

    do xi = xSize, 1, -1
      if (any(mask(xi, :))) then
        maxXi = xi
    exit
      end if
    end do

    do yi = 1, ySize
      if (any(mask(:, yi))) then
        minYi = yi
    exit
      end if
    end do

    do yi = ySize, 1, -1
      if (any(mask(:, yi))) then
        maxYi = yi
    exit
      end if
    end do

  end subroutine maskArrayLimits

  !---------------------------------------------------------------------
  integer function smallestTrueIndexBool(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! smallestTrueIndex() returns the index of the first TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, size(maskVector) + 1 is returned.

    implicit none

    logical(bool), intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestTrueIndexBool = size(maskVector) + 1
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestTrueIndexBool = i
    exit
      end if
    end do

  end function smallestTrueIndexBool

  !---------------------------------------------------------------------
  integer function smallestTrueIndexDefault(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! smallestTrueIndex() returns the index of the first TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, size(maskVector) + 1 is returned.

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    smallestTrueIndexDefault = size(maskVector) + 1
    do i = 1, size(maskVector)
    if (maskVector(i)) then
        smallestTrueIndexDefault = i
    exit
      end if
    end do

  end function smallestTrueIndexDefault

  !---------------------------------------------------------------------
  integer function largestTrueIndexBool(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! largestTrueIndex() returns the index of the last TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, 0 is returned. NOTE that the
  ! behaviour of getBinNumber() depends on this default, so don't change
  ! it!

    implicit none

    logical(bool), intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestTrueIndexBool = 0
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestTrueIndexBool = i
    exit
      end if
    end do

  end function largestTrueIndexBool

  !---------------------------------------------------------------------
  integer function largestTrueIndexDefault(maskVector)
  !---------------------------------------------------------------------
  ! The functions smallestTrueIndex() and largestTrueIndex() are related
  ! to maskArrayLimits() but are much more basic. The function
  ! largestTrueIndex() returns the index of the last TRUE value of the
  ! input logical vector maskVector(). The first index is assumed to be 1
  ! not 0.
  !
  ! If maskVector() contains no TRUE values, 0 is returned. NOTE that the
  ! behaviour of getBinNumber() depends on this default, so don't change
  ! it!

    implicit none

    logical, intent(in) :: maskVector(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    largestTrueIndexDefault = 0
    do i = size(maskVector), 1, -1
      if (maskVector(i)) then
        largestTrueIndexDefault = i
    exit
      end if
    end do

  end function largestTrueIndexDefault

  !---------------------------------------------------------------------
  function getBinNumberSingle(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(single), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestTrueIndex(logical(xBinEdges <= x, bool))

  end function getBinNumberSingle

  !---------------------------------------------------------------------
  function getBinNumberDouble(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(double), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      binNum = 0
  return
    end if

    binNum = largestTrueIndex(logical(xBinEdges <= x, bool))

  end function getBinNumberDouble

  !---------------------------------------------------------------------
  function maxLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single),  intent(in)           :: vector(:)
    logical(bool), intent(in), optional :: mask(size(vector))
    integer                             :: maxLoc1dSingle

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dSingle = maxLocVec(1)

  end function maxLoc1dSingle

  !---------------------------------------------------------------------
  function maxLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double),  intent(in)           :: vector(:)
    logical(bool), intent(in), optional :: mask(size(vector))
    integer                             :: maxLoc1dDouble

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dDouble = maxLocVec(1)

  end function maxLoc1dDouble

  !---------------------------------------------------------------------
  function minLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single),  intent(in)           :: vector(:)
    logical(bool), intent(in), optional :: mask(size(vector))
    integer                             :: minLoc1dSingle

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dSingle = minLocVec(1)

  end function minLoc1dSingle

  !---------------------------------------------------------------------
  function minLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double),  intent(in)           :: vector(:)
    logical(bool), intent(in), optional :: mask(size(vector))
    integer                             :: minLoc1dDouble

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dDouble = minLocVec(1)

  end function minLoc1dDouble


  !---------------------------------------------------------------------
  function findUniqueValuesInt8(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finshed with it!

    implicit none

    integer(int8), intent(in)           :: vector(:)
    logical,       intent(in), optional :: mask(size(vector))
    integer(int8), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int8), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt8

  !---------------------------------------------------------------------
  function findUniqueValuesInt16(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finshed with it!

    implicit none

    integer(int16), intent(in)           :: vector(:)
    logical,        intent(in), optional :: mask(size(vector))
    integer(int16), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int16), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt16

  !---------------------------------------------------------------------
  function findUniqueValuesInt32(vector, mask) result(uniqueValues)
  !---------------------------------------------------------------------
  ! Note: you should call this function by pointer assignment, viz
  !
  !	myPointer => findUniqueValues(vector)
  !
  ! Remember to then deallocate myPointer after you've finshed with it!

    implicit none

    integer(int32), intent(in)           :: vector(:)
    logical,        intent(in), optional :: mask(size(vector))
    integer(int32), pointer              :: uniqueValues(:)

    integer :: numValues, numLiveValues, i, numUniqueValues
    logical, allocatable :: localMask(:)
    integer(int32), allocatable :: tempVector(:), tempUniqueVals(:)

    !-------------------------------------------------------------------
    ! Start

    numValues = size(vector)
    allocate(localMask(numValues))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if
    
    numLiveValues = count(localMask)
    if (numLiveValues <= 0) then
      allocate(uniqueValues(0))
return
    end if

    allocate(tempVector(numLiveValues), tempUniqueVals(numLiveValues))
    tempVector = pack(vector, localMask)
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numUniqueValues = 1
    tempUniqueVals(numUniqueValues) = tempVector(1)
    do i = 2, numLiveValues
      if (tempVector(i) .ne. tempVector(i-1)) then
        numUniqueValues = numUniqueValues + 1
        tempUniqueVals(numUniqueValues) = tempVector(i)
      end if
    end do
    allocate(uniqueValues(numUniqueValues))
    uniqueValues = tempUniqueVals(:numUniqueValues)

  end function findUniqueValuesInt32


  !---------------------------------------------------------------------
  function vectorCrossSingle(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: vectorA(3), vectorB(3)
    real(single)             :: vectorCrossSingle(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossSingle(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossSingle(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossSingle(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossSingle

  !---------------------------------------------------------------------
  function vectorCrossDouble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: vectorA(3), vectorB(3)
    real(double)             :: vectorCrossDouble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDouble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDouble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDouble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDouble

  !---------------------------------------------------------------------
  function normalizeVectorSingle(argument)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: argument(:)
    real(single) :: normalizeVectorSingle(size(argument)), vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

!*** use parallel facilities!

    vectorLength2 = 0.0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0) then
      normalizeVectorSingle = 0.0
    else
      normalizeVectorSingle = argument / sqrt(vectorLength2)
    end if

  end function normalizeVectorSingle

  !---------------------------------------------------------------------
  function normalizeVectorDouble(argument)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: argument(:)
    real(double)             :: normalizeVectorDouble(size(argument))

    real(double) :: vectorLength2
    integer :: i

    !-------------------------------------------------------------------
    ! Start

!*** use parallel facilities!

    vectorLength2 = 0.0d0
    do i = 1, size(argument)
      vectorLength2 = vectorLength2 + argument(i) * argument(i)
    end do
    if (vectorLength2 .le. 0.0d0) then
      normalizeVectorDouble = 0.0d0
    else
      normalizeVectorDouble = argument / sqrt(vectorLength2)
    end if

  end function normalizeVectorDouble

  !---------------------------------------------------------------------
  subroutine findEdges(mask, figureEdges, groundEdges)
  !---------------------------------------------------------------------
  ! Given a logical array mask(), this subroutine returns an array
  ! figureEdges() which is true at all true pixels in mask() which are
  ! _not_ surrounded by nearest neighbours which are all true. The
  ! 'nearest neighbours' of a pixel at(x,y) are defined as those
  ! positions (x',y') for which x-1<=x'<=x+1 and y-1<=y'<=y+1. However,
  ! values of (x',y') that fall outside the boundaries of mask() are
  ! not counted as nearest neighbours: hence, no pixel on the boundary
  ! of mask() is set as an edge pixel just because it is on the boundary.
  !
  ! A second array groundEdges() is also returned, which is equivalent
  ! to figureEdges() calculated for .not.mask().
  !
  ! Example input mask:
  !
  !	11110000111
  !	11111000011
  !	11111100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	11101111100
  !
  ! The returned array figureEdges should look as follows:
  !	
  !	00010000110
  !	00011000010
  !	00001100011
  !	11111100001
  !	00000000000
  !	11111110000
  !	00101011100
  !
  ! The returned array groundEdges should look as follows:
  !	
  !	00001101000
  !	00000111100
  !	00000010100
  !	00000010110
  !	11111111011
  !	00000001110
  !	00010000010

    implicit none

    logical(bool), intent(in)  :: mask(:,:)
    logical(bool), intent(out), optional ::&
      figureEdges(size(mask, 1),size(mask, 2)),&
      groundEdges(size(mask, 1),size(mask, 2))

    logical(bool), allocatable :: edgeMaskBig(:,:), edgeMaskSmall(:,:)
    integer :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(mask, 1)
    ySize = size(mask, 2)

    if (.not. present(figureEdges) .and. .not. present(groundEdges)) then
  return
    end if

    allocate(edgeMaskBig(0:xSize+1,0:ySize+1), edgeMaskSmall(xSize, ySize))

    edgeMaskSmall = .false. ! default

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xSize-1, 0:ySize-1) = mask
    edgeMaskBig(0:xSize-1, 1:ySize)&
      = edgeMaskBig(0:xSize-1, 1:ySize)   .or. mask
    edgeMaskBig(0:xSize-1, 2:ySize+1)&
      = edgeMaskBig(0:xSize-1, 2:ySize+1) .or. mask
    edgeMaskBig(1:xSize,   0:ySize-1)&
      = edgeMaskBig(1:xSize  , 0:ySize-1) .or. mask
    edgeMaskBig(1:xSize,   1:ySize)&
      = edgeMaskBig(1:xSize  , 1:ySize)   .or. mask
    edgeMaskBig(1:xSize,   2:ySize+1)&
      = edgeMaskBig(1:xSize  , 2:ySize+1) .or. mask
    edgeMaskBig(2:xSize+1, 0:ySize-1)&
      = edgeMaskBig(2:xSize+1, 0:ySize-1) .or. mask
    edgeMaskBig(2:xSize+1, 1:ySize)&
      = edgeMaskBig(2:xSize+1, 1:ySize)   .or. mask
    edgeMaskBig(2:xSize+1, 2:ySize+1)&
      = edgeMaskBig(2:xSize+1, 2:ySize+1) .or. mask

    edgeMaskSmall = edgeMaskBig(1:xSize, 1:ySize)

    edgeMaskBig = .false. ! default
    edgeMaskBig(0:xSize-1, 0:ySize-1) = .not. mask
    edgeMaskBig(0:xSize-1, 1:ySize)&
      = edgeMaskBig(0:xSize-1, 1:ySize)   .or. .not. mask
    edgeMaskBig(0:xSize-1, 2:ySize+1)&
      = edgeMaskBig(0:xSize-1, 2:ySize+1) .or. .not. mask
    edgeMaskBig(1:xSize,   0:ySize-1)&
      = edgeMaskBig(1:xSize  , 0:ySize-1) .or. .not. mask
    edgeMaskBig(1:xSize,   1:ySize)&
      = edgeMaskBig(1:xSize  , 1:ySize)   .or. .not. mask
    edgeMaskBig(1:xSize,   2:ySize+1)&
      = edgeMaskBig(1:xSize  , 2:ySize+1) .or. .not. mask
    edgeMaskBig(2:xSize+1, 0:ySize-1)&
      = edgeMaskBig(2:xSize+1, 0:ySize-1) .or. .not. mask
    edgeMaskBig(2:xSize+1, 1:ySize)&
      = edgeMaskBig(2:xSize+1, 1:ySize)   .or. .not. mask
    edgeMaskBig(2:xSize+1, 2:ySize+1)&
      = edgeMaskBig(2:xSize+1, 2:ySize+1) .or. .not. mask

    edgeMaskSmall = edgeMaskSmall .and. edgeMaskBig(1:xSize, 1:ySize)

    if (present(figureEdges)) then
      figureEdges = .false.
      figureEdges = edgeMaskSmall .and. mask
    end if
    if (present(groundEdges)) then
      groundEdges = .false.
      groundEdges = edgeMaskSmall .and. .not. mask
    end if

  end subroutine findEdges

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt8(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int8), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int8), pointer :: tempSingleArray(:,:)
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt8

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt16(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int16), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int16), pointer :: tempSingleArray(:,:)
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt16

  !---------------------------------------------------------------------
  subroutine selectArraySubsetInt32(image&
  , xStartArg, xFinisArg, yStartArg, yFinisArg)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    integer(int32), pointer :: image(:,:)
    integer(int16), intent(in) :: xStartArg, xFinisArg, yStartArg, yFinisArg

    integer(int32), pointer :: tempSingleArray(:,:)
    integer(int16) :: xStart, xFinis, yStart, yFinis

    !-------------------------------------------------------------------
    ! Start

    xStart = max(lbound(image, 1), int(xStartArg))
    xFinis = min(ubound(image, 1), int(xFinisArg))
    yStart = max(lbound(image, 2), int(yStartArg))
    yFinis = min(ubound(image, 2), int(yFinisArg))

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetInt32

  !---------------------------------------------------------------------
  subroutine selectArraySubsetSingle(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    real(single), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    real(single), pointer :: tempSingleArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetSingle

  !---------------------------------------------------------------------
  subroutine selectArraySubsetDouble(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    real(double), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    real(double), pointer :: tempSingleArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetDouble

  !---------------------------------------------------------------------
  subroutine selectArraySubsetBool(image, xStart, xFinis, yStart, yFinis)
  !---------------------------------------------------------------------
  ! These subroutines are used where a pointer is being used to store an
  ! array and it is desired to resize the pointer.
  !
  ! Remember to deallocate image() after use.

    implicit none

    logical(bool), pointer :: image(:,:)
    integer(int16), intent(in) :: xStart, xFinis, yStart, yFinis

    logical(bool), pointer :: tempSingleArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempSingleArray(xStart:xFinis, yStart:yFinis))
    tempSingleArray = image(xStart:xFinis, yStart:yFinis)
    deallocate(image)
    allocate(image(xStart:xFinis, yStart:yFinis))
    image => tempSingleArray

    ! NOTE! We should NOT deallocate tempSingleArray now or ever, because
    ! it is pointing to the same area of memory as image. When we want to
    ! clear this memory, it suffices to deallocate(image).

  end subroutine selectArraySubsetBool

  !---------------------------------------------------------------------
  subroutine cyclicConvolution(vector, convolver, mask, padPlease&
  , normalise, interpolationLimit, interpolationMask, uncerts, status)
  !---------------------------------------------------------------------

    implicit none

    real,          intent(in)               :: convolver(:)
    real,          intent(inout)            :: vector(:)
    real,          intent(inout),  optional :: uncerts(:)
    logical(bool), intent(in),     optional :: mask(:), padPlease, normalise
    integer,       intent(in),     optional :: interpolationLimit
    integer,       intent(out),    optional :: status
    logical(bool), intent(out),    optional :: interpolationMask(:)

    integer :: vSize, csize, n, i, j, localStatus, psize, cHalfWidth,&
      interpolateHolesSmallerThan, holeStart, holeFinis
    real, allocatable :: paddedVector(:), output(:)
    real :: summ, weight, norm
    logical :: localPadPlease, localNormalise, inNullPatch
    logical(bool), allocatable :: localMask(:), localInterpolationMask(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    vSize = size(vector)
    csize = size(convolver)
    cHalfWidth = csize / 2

    if (present(padPlease)) then
      localPadPlease = padPlease
    else
      localPadPlease = .true.
    end if
    if (present(normalise)) then
      localNormalise = normalise
    else
      localNormalise = .true.
    end if
    if (present(interpolationLimit)) then
      interpolateHolesSmallerThan = interpolationLimit
    else
      interpolateHolesSmallerThan = 0
    end if

    if (interpolateHolesSmallerThan .gt. csize) then
      if (present(status)) status = -1
  return
    end if

    if (localPadPlease) then
      psize = vSize + csize ! Pad doesn't need to be quite this big, but
      ! it won't hurt.
    else
      psize = vSize
    end if

    if (localNormalise) then
      norm = 1.0
    else
      norm = sum(convolver)
    end if

    allocate(paddedVector(psize), output(psize), localMask(psize)&
    , localInterpolationMask(psize))

    localMask = .false.
    if (present(mask)) then
      if (size(mask) .ne. vSize) then
        if (present(status)) status = -3
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      localMask(1:vSize) = mask

    else
      localMask(1:vSize) = .true.
    end if

    ! Check all the holes to see if any are small enough to be interpolated:
    localInterpolationMask = localMask ! vector default
    if (.not. localMask(1)) then
      inNullPatch = .true.
      holeStart = 1
    else
      inNullPatch = .false.
    end if
    do i = 2, vSize
      if (inNullPatch .and. localMask(i)) then
        holeFinis = i - 1
        if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
          localInterpolationMask(holeStart:holeFinis) = .true.
        end if
        inNullPatch = .false.

      else if (.not. inNullPatch .and. .not. localMask(i)) then
        holeStart = i
        inNullPatch = .true.
      end if
    end do
    if (inNullPatch) then
      holeFinis = vSize
      if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
        localInterpolationMask(holeStart:holeFinis) = .true.
      end if
    end if

    paddedVector(1:vSize) = vector
    output = 0.0 ! vector default.
    do n = 1, vSize
    if (.not. localInterpolationMask(n)) cycle
      summ = 0.0
      weight = 0.0
      do i = 1, csize
        j = n - cHalfWidth + i - 1
        if (j .le. 0)     j = j + psize
        if (j .gt. psize) j = j - psize
      if (.not. localMask(j)) cycle
        summ = summ + convolver(i) * paddedVector(j)
        weight = weight + convolver(i)
      end do
      if (weight .le. 0.0)  then
        if (present(status)) status = -4
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      output(n) = summ * norm / weight
    end do

    vector = output(1:vSize)

    if (present(uncerts)) then
      paddedVector(1:vSize) = uncerts * uncerts
      output = 0.0 ! vector default.
      do n = 1, vSize
      if (.not. localInterpolationMask(n)) cycle
        summ = 0.0
        weight = 0.0
        do i = 1, csize
          j = n - cHalfWidth + i - 1
          if (j .le. 0)     j = j + psize
          if (j .gt. psize) j = j - psize
        if (.not. localMask(j)) cycle
          summ = summ + convolver(i) * convolver(i) * paddedVector(j)
          weight = weight + convolver(i)
        end do
        if (weight .le. 0.0)  then
          output(n) = 0.0
        else
          output(n) = sqrt(summ) * norm / weight
        end if
      end do

      uncerts = output(1:vSize)
    end if

    if (present(interpolationMask))&
      interpolationMask = localInterpolationMask(1:vSize)

    deallocate(paddedVector, output, localMask, localInterpolationMask)

    if (present(status)) status = localStatus

  end subroutine cyclicConvolution

  !---------------------------------------------------------------------
  function rotateArraySingle(inArray, numQuarterRevs) result(outArray)
  !---------------------------------------------------------------------
  ! Rotates the array clockwise by numQuarterRevs steps. The array must be square - if not, it is returned unchanged.

!*** NOT YET TESTED!!

    implicit none

    real(single), intent(in) :: inArray(:,:)
    integer, intent(in) :: numQuarterRevs
    real(single) :: outArray(size(inArray,1),size(inArray,2))

    integer :: aSize

    !-------------------------------------------------------------------
    ! Start

    aSize = size(inArray, 1)
    if (size(inArray, 2) .ne. aSize .or. aSize <= 0) then
      outArray = inArray
  return
    end if

    select case(modulo(numQuarterRevs, 4))
    case(0)
      outArray = inArray
    case(1)
      outArray(:,aSize:1:-1) = transpose(inArray)
    case(2)
      outArray(aSize:1:-1,aSize:1:-1) = inArray
    case(3)
      outArray(aSize:1:-1,:) = transpose(inArray)
    end select

  end function rotateArraySingle

  !---------------------------------------------------------------------
  function linearInterpolate(x, y, xSample) result(ySample)
  !---------------------------------------------------------------------
  ! Note that x values must be monotonically increasing.

    implicit none

    real(single), intent(in) :: x(:), y(size(x)), xSample
    real(single) :: ySample

    integer :: numSamples, bin

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(x)

    if (numSamples <= 0) then
      ySample = 0.0
  return
    end if

    if (numSamples == 1) then
      if (xSample == x(1)) then
        ySample = y(1)
      else
        ySample = 0.0
      end if
  return
    end if

    if (any(x(:numSamples-1) > x(2:))) then
      ySample = 0.0
  return
    end if

    if (xSample < x(1) .or. xSample > x(numSamples)) then
      ySample = 0.0
  return
    end if

    bin = getBinNumber(x, xSample)
    ySample = y(bin) + (xSample - x(bin)) * (y(bin+1) - y(bin))&
      / (x(bin+1) - x(bin))

  end function linearInterpolate

  !---------------------------------------------------------------------
  function trapezoidSumSingle(x, y) result(summ)
  !---------------------------------------------------------------------
  ! Given two vectors x() and y() containing samples of a function F, this
  ! subroutine returns the 'trapezoidal-rule' estimate of the integral of F.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    real(single)             :: summ

    integer(int32) :: i, n

    !-------------------------------------------------------------------
    ! Start

    summ = 0.0 ! default
    n = size(x)
  if (n <= 1) return

    do i = 1, n - 1
      summ = summ + x(i + 1) * y(i) - x(i) * y(i + 1)
    end do
    summ = 0.5 * (x(n) * y(n) - x(1) * y(1) + summ)

  end function trapezoidSumSingle

  !---------------------------------------------------------------------
  function trapezoidSumDouble(x, y) result(summ)
  !---------------------------------------------------------------------
  ! Given two vectors x() and y() containing samples of a function F, this
  ! subroutine returns the 'trapezoidal-rule' estimate of the integral of F.

    implicit none

    real(double), intent(in) :: x(:), y(size(x))
    real(double)             :: summ

    integer(int32) :: i, n

    !-------------------------------------------------------------------
    ! Start

    summ = 0.0d0 ! default
    n = size(x)
  if (n <= 1) return

    do i = 1, n - 1
      summ = summ + x(i + 1) * y(i) - x(i) * y(i + 1)
    end do
    summ = 0.5 * (x(n) * y(n) - x(1) * y(1) + summ)

  end function trapezoidSumDouble

  !---------------------------------------------------------------------
  subroutine getGridDiffs(arrayOfOffsets, xi, yi, xDiff, yDiff, status)
  !---------------------------------------------------------------------
  ! This and the sub interpolateGridDiff() are designed to cope with interpolations within a 2d grid, for each point of which there is a defined transformed location, ie a location on another grid. The position of the interpolated point on the second grid can be estimated using (i) the positional on the second grid of a base point on the first, and (ii) the differential relations between the two grids at that point.
  !
  ! The task of the present subroutine is, given the two grids and a position on one, to estimate these differentials.

    implicit none

    type(Point2dT), intent(in)            :: arrayOfOffsets(:,:)
    integer,        intent(in)            :: xi, yi
    type(Point2dT), intent(out)           :: xDiff, yDiff
    integer,        intent(out), optional :: status
    ! Point2dT type defined in geometric_types.

    integer :: xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xDiff%x = 0.0
    xDiff%y = 0.0
    yDiff%x = 0.0
    yDiff%y = 0.0

    xSize = size(arrayOfOffsets, 1)
    ySize = size(arrayOfOffsets, 2)
    if (xSize < 2 .or. ySize < 2) then
      if (present(status)) status = -1
  return
    end if
    if (xi < 1 .or. xi > xSize .or. yi < 1 .or. yi > ySize) then
      if (present(status)) status = -2
  return
    end if

    if (xi == 1) then
      xDiff%x = arrayOfOffsets(2, yi)%x - arrayOfOffsets(1, yi)%x
      xDiff%y = arrayOfOffsets(2, yi)%y - arrayOfOffsets(1, yi)%y
    else if (xi == xSize) then
      xDiff%x = arrayOfOffsets(xSize, yi)%x - arrayOfOffsets(xSize - 1, yi)%x
      xDiff%y = arrayOfOffsets(xSize, yi)%y - arrayOfOffsets(xSize - 1, yi)%y
    else
      xDiff%x = (arrayOfOffsets(xi + 1, yi)%x - arrayOfOffsets(xi - 1, yi)%x)&
        / 2.0
      xDiff%y = (arrayOfOffsets(xi + 1, yi)%y - arrayOfOffsets(xi - 1, yi)%y)&
        / 2.0
    end if

    if (yi == 1) then
      yDiff%x = arrayOfOffsets(xi, 2)%x - arrayOfOffsets(xi, 1)%x
      yDiff%y = arrayOfOffsets(xi, 2)%y - arrayOfOffsets(xi, 1)%y
    else if (yi == ySize) then
      yDiff%x = arrayOfOffsets(xi, ySize)%x - arrayOfOffsets(xi, ySize - 1)%x
      yDiff%y = arrayOfOffsets(xi, ySize)%y - arrayOfOffsets(xi, ySize - 1)%y
    else
      yDiff%x = (arrayOfOffsets(xi, yi + 1)%x - arrayOfOffsets(xi, yi - 1)%x)&
        / 2.0
      yDiff%y = (arrayOfOffsets(xi, yi + 1)%y - arrayOfOffsets(xi, yi - 1)%y)&
        / 2.0
    end if

    if (present(status)) status = 0

  end subroutine getGridDiffs

  !---------------------------------------------------------------------
  subroutine interpolateGridDiff(offset, xDiff, yDiff, fracVector, fracOffsetVector)
  !---------------------------------------------------------------------
  ! This and the sub interpolateGridDiff() are designed to cope with interpolations within a 2d grid, for each point of which there is a defined transformed location, ie a location on another grid. The position of the interpolated point on the second grid can be estimated using (i) the positional on the second grid of a base point on the first, and (ii) the differential relations between the two grids at that point.
  !
  ! The task of the present subroutine is to perform the interpolation.

    implicit none

    type(Point2dT), intent(in) :: offset, xDiff, yDiff, fracVector(:)
    type(Point2dT), intent(out) :: fracOffsetVector(size(fracVector))
    ! Point2dT type defined in geometric_types.

    !-------------------------------------------------------------------
    ! Start

    fracOffsetVector%x = offset%x + fracVector%x&
      * xDiff%x + fracVector%y * yDiff%x
    fracOffsetVector%y = offset%y + fracVector%x&
      * xDiff%y + fracVector%y * yDiff%y

  end subroutine interpolateGridDiff


!  !---------------------------------------------------------------------
!  real(single) function cohmMedianInt32Vector(vector, maskVector)&
!  result(cohmMedian)
!  !---------------------------------------------------------------------
!  ! A histogram h(n) as a function of integer n is first made of the values in the input integer vector. The result r is then calculated such that the sum of the histogram values for n<r equals the sum for n>r. For n such that n<=r and (n+1) > r, the values in this nth histogram bin are divided, h(n)*(r-n) being added to the 'lefthand' weights, h(n)*(n+1-r) to the 'righthand'. This is approximately the same as a median where typical values of n are large. However the bin-splitting property makes it a useful extension of the median concept to cases where most n are small (ie of order 1).
!
!    implicit none
!
!    integer(int32), intent(in)           :: vector(:)
!    logical,        intent(in), optional :: maskVector(:)
!
!    real(single)                         :: cohmMedian
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    call makeHistogram()
!
!    medianBin = medianInt32Vector(vector, maskVector)
!    lhArea = 
!    do while()
!      medianBin = medianBin + 1
!    end do
!
!  end function cohmMedianInt32Vector


  !---------------------------------------------------------------------
  real(single) function medianSingleArray(array, maskArray)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(array,1)*size(array,2) the function
  ! returns the value in the lower central bin of the sorted vector, rather
  ! than the (more correct) mean of the lower and upper central bins. For
  ! ODD-numbered size the function returns the value in the central bin of
  ! the sorted vector, as expected.

    implicit none

    real(single), intent(in)           :: array(:,:)
    logical,      intent(in), optional :: maskArray(:,:)

    integer :: vectorShape(1)

    !-------------------------------------------------------------------
    ! Start

    vectorShape(1) = size(array, 1) * size(array, 2)
    if (present(maskArray)) then
      medianSingleArray = valueAtHistoFracSingleVec(reshape(array, vectorShape)&
        , 0.5, reshape(maskArray, vectorShape))
    else
      medianSingleArray = valueAtHistoFracSingleVec(reshape(array, vectorShape)&
        , 0.5)
    end if

  end function medianSingleArray

  !---------------------------------------------------------------------
  real(double) function medianDoubleArray(array, maskArray)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(array,1)*size(array,2) the function
  ! returns the value in the lower central bin of the sorted vector, rather
  ! than the (more correct) mean of the lower and upper central bins. For
  ! ODD-numbered size the function returns the value in the central bin of
  ! the sorted vector, as expected.

    implicit none

    real(double), intent(in)           :: array(:,:)
    logical,      intent(in), optional :: maskArray(:,:)

    integer :: vectorShape(1)

    !-------------------------------------------------------------------
    ! Start

    vectorShape(1) = size(array, 1) * size(array, 2)
    if (present(maskArray)) then
      medianDoubleArray = valueAtHistoFracDoubleVec(reshape(array, vectorShape)&
        , 0.5d0, reshape(maskArray, vectorShape))
    else
      medianDoubleArray = valueAtHistoFracDoubleVec(reshape(array, vectorShape)&
        , 0.5d0)
    end if

  end function medianDoubleArray

  !---------------------------------------------------------------------
  real(single) function medianSingleVector(vector, maskVector)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(vector) the function returns the value
  ! in the lower central bin of the sorted vector, rather than the (more
  ! correct) mean of the lower and upper central bins. For ODD-numbered size
  ! the function returns the value in the central bin of the sorted vector,
  ! as expected.

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    medianSingleVector = valueAtHistoFracSingleVec(vector, 0.5, maskVector)

  end function medianSingleVector

  !---------------------------------------------------------------------
  real(double) function medianDoubleVector(vector, maskVector)
  !---------------------------------------------------------------------
  ! NOTE that for EVEN-numbered size(vector) the function returns the value
  ! in the lower central bin of the sorted vector, rather than the (more
  ! correct) mean of the lower and upper central bins. For ODD-numbered size
  ! the function returns the value in the central bin of the sorted vector,
  ! as expected.

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    medianDoubleVector = valueAtHistoFracDoubleVec(vector, 0.5d0, maskVector)

  end function medianDoubleVector

  !---------------------------------------------------------------------
  real(single) function valueAtHistoFracSingleVec(vector, binFraction&
    , maskVector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: vector(:), binFraction
    logical,      intent(in), optional :: maskVector(:)

    !-------------------------------------------------------------------
    ! Start

    valueAtHistoFracSingleVec = real(valueAtHistoFracDoubleVec(&
      real(vector, double), real(binFraction, double), maskVector), single)

  end function valueAtHistoFracSingleVec

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFracDoubleVec(vector, binFraction&
    , maskVector)
  !---------------------------------------------------------------------
  ! This function takes a vector of real values, sorts it in increasing
  ! order, and returns that element whose bin index (starting at 1) in
  ! the sorted vector, when divided by the total number of bins in the
  ! vector, is closest to the argument binFraction. Eg to find the median
  ! value of a vector one could call this function with binFraction set
  ! to 0.5.

!****maskVector mods not yet tested

    implicit none

    real(double), intent(in)           :: vector(:), binFraction
    logical,      intent(in), optional :: maskVector(:)

    integer :: nearestBin, numBins
    real(double), allocatable :: repackedVector(:)
    real(double) :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    adjustedBinFraction = min(1.0d0, max(0.0d0, binFraction))

    if (present(maskVector)) then
      numBins = count(maskVector)
    else
      numBins = size(vector)
    end if

    if (adjustedBinFraction > 0.5) then
      nearestBin = 1 + numBins - ceiling(numBins * (1.0 - adjustedBinFraction))
    else
      nearestBin = ceiling(numBins * adjustedBinFraction)
    end if
    nearestBin = min(numBins, max(1, nearestBin))

    if (present(maskVector)) then
      allocate(repackedVector(count(maskVector)))
      repackedVector = pack(vector, maskVector)
      valueAtHistoFracDoubleVec = valueAtHistoFractionKernel(repackedVector&
        , nearestBin)

    else
      valueAtHistoFracDoubleVec = valueAtHistoFractionKernel(vector&
        , nearestBin)
    end if

  end function valueAtHistoFracDoubleVec

  !---------------------------------------------------------------------
  real(single) function valueAtHistoFracSingleArr(array, binFraction, maskArray)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: array(:,:), binFraction
    logical,      intent(in), optional :: maskArray(size(array,1)&
                                                   ,size(array,2))


    !-------------------------------------------------------------------
    ! Start

    valueAtHistoFracSingleArr = real(valueAtHistoFracDoubleArr(&
      real(array, double), real(binFraction, double), maskArray), single)

  end function valueAtHistoFracSingleArr

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFracDoubleArr(array, binFraction, maskArray)
  !---------------------------------------------------------------------
  ! This function takes an array of real values, reshapes it to a vector, sorts the vector in increasing order, and returns the value of that element whose bin index (starting at 1) in the sorted vector, when divided by the total number of bins in the vector, is closest to the argument binFraction. Eg to find the median value of an array one could call this function with binFraction set to 0.5.
  !
  ! If maskArray() is supplied the non-masked pixels are ignored in the calculation.

!*** mods from _20030817 not yet tested

    implicit none

    real(double), intent(in)           :: array(:,:), binFraction
    logical,      intent(in), optional :: maskArray(size(array,1)&
                                                   ,size(array,2))

    integer :: xSize, ySize, vSize, nearestBin, numBins
    real(double), allocatable :: vector(:)
    real(double) :: adjustedBinFraction

    !-------------------------------------------------------------------
    ! Start

    xSize = size(array, 1)
    ySize = size(array, 2)
    vSize = xSize * ySize

    if (present(maskArray)) then
      allocate(vector(count(maskArray)))
      vector = pack(array, maskArray)
      numBins = count(maskArray)

    else
      allocate(vector(vSize))
      vector = reshape(array, (/vSize/))
      numBins = vSize
    end if

    adjustedBinFraction = min(1.0d0, max(0.0d0, binFraction))
    if (adjustedBinFraction > 0.5) then
      nearestBin = 1 + numBins - ceiling(numBins * (1.0 - adjustedBinFraction))
    else
      nearestBin = ceiling(numBins * adjustedBinFraction)
    end if
    nearestBin = min(numBins, max(1, nearestBin))

    valueAtHistoFracDoubleArr = valueAtHistoFractionKernel(vector, nearestBin)

  end function valueAtHistoFracDoubleArr

  !---------------------------------------------------------------------
  real(double) function valueAtHistoFractionKernel(vector, nearestBin)
  !---------------------------------------------------------------------
  ! Adapted from the Numerical Recipes function select().

    implicit none

    real(double), intent(in) :: vector(:)
    integer, intent(in) :: nearestBin

    integer :: vSize, i, ir, j, l, mid
    real(double) :: tempVal
    real(double), allocatable :: tempVec(:)

    !-------------------------------------------------------------------
    ! Start

    vSize = size(vector)

    if (vSize <= 0) then
      valueAtHistoFractionKernel = 0.0d0
  return
    end if

    allocate(tempVec(vSize))
    tempVec = vector

    l = 1
    ir = vSize

    do
      if (ir - l <= 1) then
        if (ir - l == 1) then
          if (tempVec(l) > tempVec(ir)) call swap(tempVec(l), tempVec(ir))
          ! swap() is in ssc_misc.f90
        end if
        valueAtHistoFractionKernel = tempVec(nearestBin)
    exit

      else
        mid = (l + ir) / 2
        call swap(tempVec(mid), tempVec(l + 1))

        if (tempVec(l)   > tempVec(ir))  call swap(tempVec(l),   tempVec(ir))
        if (tempVec(l+1) > tempVec(ir))  call swap(tempVec(l+1), tempVec(ir))
        if (tempVec(l)   > tempVec(l+1)) call swap(tempVec(l),   tempVec(l+1))
        ! swap() is in ssc_misc.f90

        i = l + 1
        j = ir
        tempVal = tempVec(l + 1)
        do
          do
            i = i + 1
          if (tempVec(i) >= tempVal) exit
          end do

          do
            j = j - 1
          if (tempVec(j) <= tempVal) exit
          end do

        if (j < i) exit

          call swap(tempVec(i), tempVec(j))
          ! swap() is in ssc_misc.f90
        end do

        tempVec(l + 1) = tempVec(j)
        tempVec(j) = tempVal
        if (j >= nearestBin) ir = j - 1
        if (j <= nearestBin) l = i
      end if
    end do

  end function valueAtHistoFractionKernel

  !---------------------------------------------------------------------
  real(single) function angleBetweenSingle(vectorA, vectorB)
  !---------------------------------------------------------------------

!****** check that these give a more precise value than the cosine rule!! It's by no means obvious.


  ! As is well known, the cosine of the angle between two vectors a_ and
  ! b_ is just (a_.b_)/|a_|/|b_|. Direct inversion of this is however not
  ! very precise when a_ and b_ are nearly parallel. Better is to also
  ! find the sine, then use atan2 to get the angle. This does incur a
  ! penalty however, since it involves calculation of 2 square roots.
  !
  ! The sine can be shown to equal
  !
  !	              |   b_       a_ . b_      a_   |
  !	sine(angle) = | ------ - ---------- * ------ |.
  !	              |  |b_|     |a_||b_|     |a_|  |
  !
  ! However it is more convenient to calculate |a_|*|b_|*cos and
  ! |a_|*|b_|*sin instead of cos and sin.

    implicit none

    real(single), intent(in) :: vectorA(:), vectorB(size(vectorA))

    real(single) :: normA, abCos, abSinVector(size(vectorA)), abSin

    !-------------------------------------------------------------------
    ! Start

    normA = sqrt(sum(vectorA * vectorA))
    abCos = sum(vectorA * vectorB)
    abSinVector = vectorB * normA - abCos * vectorA / normA
    abSin = sqrt(sum(abSinVector * abSinVector))

    angleBetweenSingle = atan2(abSin, abCos)

  end function angleBetweenSingle

  !---------------------------------------------------------------------
  real(double) function angleBetweenDouble(vectorA, vectorB)
  !---------------------------------------------------------------------
  ! As is well known, the cosine of the angle between two vectors a_ and
  ! b_ is just (a_.b_)/|a_|/|b_|. Direct inversion of this is however not
  ! very precise when a_ and b_ are nearly parallel. Better is to also
  ! find the sine, then use atan2 to get the angle. This does incur a
  ! penalty however, since it involves calculation of 2 square roots.
  !
  ! The sine can be shown to equal
  !
  !	              |   b_       a_ . b_      a_   |
  !	sine(angle) = | ------ - ---------- * ------ |.
  !	              |  |b_|     |a_||b_|     |a_|  |
  !
  ! However it is more convenient to calculate |a_|*|b_|*cos and
  ! |a_|*|b_|*sin instead of cos and sin.

    implicit none

    real(double), intent(in) :: vectorA(:), vectorB(size(vectorA))

    real(double) :: normA, abCos, abSinVector(size(vectorA)), abSin

    !-------------------------------------------------------------------
    ! Start

    normA = sqrt(sum(vectorA * vectorA))
    abCos = sum(vectorA * vectorB)
    abSinVector = vectorB * normA - abCos * vectorA / normA
    abSin = sqrt(sum(abSinVector * abSinVector))

    angleBetweenDouble = atan2(abSin, abCos)

  end function angleBetweenDouble

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module array_utils
