!
! Module: interpolate
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to interpolate on grids.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module interpolate

  use types

  use array_utils	! for getBinNumber
  use ssc_misc		! for fracVal.
  use splines		! for calcCubicSplines

  implicit none 

  private 

  public bilinearSample
  public bilinearValue
  public interpolateInPolarCoord

contains 

  !---------------------------------------------------------------------
  real function bilinearSample(zGridSamples, xSamples, ySamples, x, y&
  , pixelIsNull, returnedNullValue)
  !---------------------------------------------------------------------

    implicit none

    real(single),  intent(in) ::&
      xSamples(:),&
      ySamples(:),&
      zGridSamples(size(xSamples), size(ySamples)),&
      x,&
      y
    logical(bool), intent(in), optional :: pixelIsNull(size(xSamples)-1&
                                                     , size(ySamples)-1)
    real(single),  intent(in), optional :: returnedNullValue

    real(single) :: fracX, fracY
    integer :: xi, yi, xsize, ysize

    !-------------------------------------------------------------------
    ! Start

    ! Load default value:
    if (present(returnedNullValue)) then
      bilinearSample = returnedNullValue
    else
      bilinearSample = 0.0
    end if

    xsize = size(xSamples)
    ysize = size(ySamples)
  if (xsize <= 1 .or. ysize <= 1) return

    if ( x < xSamples(1) .or. x > xSamples(xsize)&
    .or. y < ySamples(1) .or. y > ySamples(ysize))&
  return

    xi = getBinNumber(xSamples, x)
    yi = getBinNumber(ySamples, y)

    if (present(pixelIsNull)) then
  if (pixelIsNull(xi, yi)) return
    end if

    fracX = (x - xSamples(xi)) / (xSamples(xi+1) - xSamples(xi))
    fracY = (y - ySamples(yi)) / (ySamples(yi+1) - ySamples(yi))
    bilinearSample = bilinearValue(zGridSamples(xi:xi+1, yi:yi+1)&
      , fracX, fracY)

  end function bilinearSample

  !---------------------------------------------------------------------
  real function bilinearValue(zAtCorners, fracX, fracY)
  !---------------------------------------------------------------------

!**** NOTE the z indices are the right way around here!! Not elsewhere.

    implicit none

    real(single), intent(in) :: zAtCorners(2,2), fracX, fracY

    !-------------------------------------------------------------------
    ! Start

    bilinearValue = (1.0 - fracX) * (1.0 - fracY) * zAtCorners(1, 1)&
             +             fracX  * (1.0 - fracY) * zAtCorners(2, 1)&
             +             fracX         * fracY  * zAtCorners(2, 2)&
             +      (1.0 - fracX)        * fracY  * zAtCorners(1, 2)

  end function bilinearValue

  !---------------------------------------------------------------------
  subroutine interpolateInPolarCoord(thetaSamplesVector, phiSamplesVector&
  , zSamplesArray, thetaMap, phiMap, interpolatedZMap)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      thetaSamplesVector(:), phiSamplesVector(:),&
      zSamplesArray(size(thetaSamplesVector),size(phiSamplesVector))
    real(double), intent(in) ::&
      thetaMap(:,:), phiMap(size(thetaMap, 1), size(thetaMap, 2))
    real(single), intent(out) ::&
      interpolatedZMap(size(thetaMap, 1), size(thetaMap, 2))

    integer(int16) ::&
      numThetaSamples, numPhiSamples,&
      xsize, ysize, loPhiI, hiPhiI,&
      xi, yi, phiI, i, j
    real(single), allocatable ::&
      secondDerivWrtTheta(:,:)
    real(single) :: thetaArcsec, phi, loPhiZValue, hiPhiZValue, phiFrac

    !-------------------------------------------------------------------
    ! Start

    numThetaSamples = size(thetaSamplesVector)
    numPhiSamples   = size(phiSamplesVector)
    ! NOTE these all must be >= 2.

    xsize = size(thetaMap, 1)
    ysize = size(thetaMap, 2)

    if  (minval(thetaMap) < thetaSamplesVector(1)) then
print *, 'minval(thetaMap), thetaSamplesVector(1): ', minval(thetaMap), thetaSamplesVector(1)
call error('bug','')
    end if

    if  (minval(phiMap) < phiSamplesVector(1)&
    .or. maxval(phiMap) > phiSamplesVector(numPhiSamples)) then
print *, 'minval(phiMap), phiSamplesVector(1): ', minval(phiMap), phiSamplesVector(1)
print *, 'maxval(phiMap), phiSamplesVector(n): ', maxval(phiMap), phiSamplesVector(numThetaSamples)
call error('bug','')
    end if

    loPhiI =  largestTrueIndex(logical(phiSamplesVector <= minval(phiMap),bool))
    hiPhiI = smallestTrueIndex(logical(phiSamplesVector >= maxval(phiMap),bool))
    ! These are in ssclib/array_utils.

    if (loPhiI == hiPhiI) then ! unlikely as it implies that
      ! there is only 1 value of phi among the image pixels, and
      ! also that that value lies exactly on one of the samples
      ! in phiSamplesVector. Nevertheless, could happen I suppose.

      if (hiPhiI < numPhiSamples) then
        hiPhiI = hiPhiI + 1
      else
        loPhiI = loPhiI - 1
      end if
    end if

    ! Preload the 2nd derivatives wrt theta for the cubic spline:

    allocate(secondDerivWrtTheta(numThetaSamples, loPhiI:hiPhiI))
    do phiI = loPhiI, hiPhiI
      call calcCubicSplines(thetaSamplesVector, zSamplesArray(:, phiI)&
        , loBoundType='fixed1st', loBoundValue=0.0&
        , hiBoundType='fixed2nd', hiBoundValue=0.0& ! only an approx,
                                                    ! don't know the
                                                    ! correct outer BC.
        , y2=secondDerivWrtTheta(:, phiI)&
      )
    end do

    ! Now interpolate the z samples array for all image pixels:

    interpolatedZMap = 0.0
    do xi = 1, xsize
      do yi = 1, ysize
        thetaArcsec = thetaMap(xi, yi)	! type conversion from
        phi = phiMap(xi, yi)		! double to single.
      if (thetaArcsec > thetaSamplesVector(numThetaSamples)) cycle

        ! Get the indices of the lower LH corner of the theta-phi
        ! rectangle in theta-phi space that contains both thetaArcsec
        ! and phi:

        i = largestTrueIndex(logical(thetaSamplesVector <= thetaArcsec, bool))
        if (i == numThetaSamples) i = numThetaSamples - 1
        j = largestTrueIndex(logical(phiSamplesVector <= phi, bool))
        if (j == hiPhiI) j = hiPhiI - 1

        ! Get the cubic-spline interpolation at each of the two
        ! bracketting phi values:

        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
          , zSamplesArray(i:i+1, j)&
          , secondDerivWrtTheta(i:i+1, j),   thetaArcsec, loPhiZValue)
        call cubicSplineInterpolate(thetaSamplesVector(i:i+1)&
          , zSamplesArray(i:i+1, j+1)&
          , secondDerivWrtTheta(i:i+1, j+1), thetaArcsec, hiPhiZValue)

        ! Finally do a linear interpolation between these two phi values:

        phiFrac = (phi - phiSamplesVector(j))&
          / (phiSamplesVector(j+1) - phiSamplesVector(j))
        interpolatedZMap(xi, yi) = fracVal(loPhiZValue, hiPhiZValue, phiFrac)
      end do
    end do

    deallocate(secondDerivWrtTheta)

    if (any(thetaMap > thetaSamplesVector(numThetaSamples)))&
      call warning('thetaOutsideRange','')

  end subroutine interpolateInPolarCoord

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module interpolate
