!
! Module: intervals_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This module contains routines used to manipulate time intervals.
! 
! Comments: 
!*** include initializeIntervalT, initializeGtiVecType
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module intervals_aux

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile:

  use caltypes	! for IntervalT
  use types
  use dss

  use dss_ranges_aux	! for orRanges, andRanges

  implicit none 

  private 

  public deallocateGtiVec
  public intervalsAreWellFormed
  public getBinIntervals
  public andIntervals
!public andIntervalsPointer
!public orIntervalsPointer
  public orIntervals
  public isWithinInterval
!  public checkIntervalOrder

  interface andIntervals
    module procedure andIntervalsScalVec,    andIntervalsScalScalVec
    module procedure andIntervalsScalVecVec, andIntervalsVecScalVec
    module procedure andIntervalsVecVec,     andIntervalsVecVecVec
    module procedure andIntervalsScalScal,   andIntervalsScalScalScal
  end interface

  interface orIntervals
    module procedure orIntervalsScalVec,    orIntervalsScalScalVec
    module procedure orIntervalsScalVecVec, orIntervalsVecScalVec
    module procedure orIntervalsVecVec,     orIntervalsVecVecVec
    module procedure orIntervalsScalScal,   orIntervalsScalScalScal
  end interface

  type, public :: gtiVecType
    type(IntervalT), pointer :: gtis(:)=>null()
    integer                  :: gtiType=0
  end type gtiVecType

  integer, parameter, public :: OPEN_GTI=0, CLOSED_GTI=1, GTI_IN_LIST=2
  ! Possible values of gtiType.

contains 

!  !---------------------------------------------------------------------
!  subroutine local_splitSetTabName(setTabName, setName, tabName, noColonFound)
!!**** local because I don't want this module to depend on dal_aux yet.
!  !---------------------------------------------------------------------
!  ! The sas param interface allows one to specify a dataset and table name
!  ! in a single parameter, by using parameter type 'table'. The format of
!  ! such a command-line argument is as in the followin example:
!  !
!  !	my_tab_parameter = fred.ds:BIGTAB
!  !
!  ! In practice it is easy inadvertently to leave off the colon + table
!  ! name. In this case the dal does not offer any very helpful message.
!  ! The aim of the present subroutine is (i) to allow people to leave off
!  ! the tab name if they choose, in which case they get pointed to the
!  ! first table in the dataset, and (ii) to issue a detailed and informatory
!  ! warning in the case that it was necessary to specify the table name but
!  ! the person forgot to do it.
!
!    implicit none
!
!    character(*),  intent(in)            :: setTabName
!    character(*),  intent(out)           :: setName, tabName
!    logical(bool), intent(out), optional :: noColonFound
!
!    integer(int16) :: posOfColon
!    type(DataSetT) :: set
!    logical :: fileExists
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    posOfColon = index(setTabName, ':')
!    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
!    ! has forgotten to include the table name.
!
!      inquire(file=setTabName, exist=fileExists)
!
!      if (.not. fileExists) then
!call error('missingFile', "No colon+block name was found after the set &
!&name. However, no dataset of that name exists, so I can't tell what block &
!&of it you meant to access.")
!      end if
!
!      call warning('missingBlockName', "No colon+block name was &
!        &found after the set name. I'm going to assume that you meant &
!        &to access the first block in the set. If that's what you wanted, &
!        &then there is no cause for alarm.")
!
!      if (present(noColonFound)) noColonFound = .true.
!
!      setName = setTabName
!      set = dataSet(setName, READ)
!      tabName = name(block(set, 0))
!      call release(set)
!
!    else
!      if (present(noColonFound)) noColonFound = .false.
!
!      setName = setTabName(:posOfColon - 1)
!      tabName = setTabName(posOfColon + 1:)
!    end if
!
!!  end subroutine local_splitSetTabName

  !---------------------------------------------------------------------
  subroutine deallocateGtiVec(gtiVec)
  !---------------------------------------------------------------------

    implicit none

    type(gtiVecType), pointer :: gtiVec(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(gtiVec)
    if (.not. associated(gtiVec(i)%gtis)) cycle
      deallocate(gtiVec(i)%gtis)
    end do
    deallocate(gtiVec)

  end subroutine deallocateGtiVec

  !---------------------------------------------------------------------
  logical(bool) function intervalsAreWellFormed(intervals)
  !---------------------------------------------------------------------
  ! This checks that the intervals vector (of size N) obeys the following
  ! conditions:
  !	(i) For all i in [1:N], intervals(i)%lower must < intervals(i)%upper. 
  !	(ii) For all i in [1:N-1], intervals(i)%upper < intervals(i+1)%lower.

!*** write a correctIntervals subroutine in analogy to correctRanges()!

    implicit none

    type(IntervalT), intent(in) :: intervals(:)

    integer :: numIntervals

    !-------------------------------------------------------------------
    ! Start

    numIntervals = size(intervals)
    if (numIntervals <= 0) then
      intervalsAreWellFormed = .true.
  return
    end if

    intervalsAreWellFormed = .false. ! default

    if (any(intervals%lower >= intervals%upper)&
    .or. any(intervals(:numIntervals-1)%upper&
    >= intervals(2:numIntervals)%lower)) then
  return
    end if

  end function intervalsAreWellFormed

  !---------------------------------------------------------------------
  function getBinIntervals(binStartTime, binStopTime, intervals)&
  result(binIntervals)
  !---------------------------------------------------------------------
  ! This function returns an interval-type vector which contains the AND of
  ! the input bin duration and the input intervals. It makes use of the
  ! range-ANDing function andRangesPair() in dssaux.f90 to compute this.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.

    implicit none

    real(double),       intent(in) :: binStartTime, binStopTime
    type(IntervalT), intent(in) :: intervals(:)
    type(IntervalT), pointer    :: binIntervals(:)

    type(RangeT) :: binTimeRange(1), timeRanges(size(intervals))
    type(RangeT), pointer :: binTimeRanges(:)

    !-------------------------------------------------------------------
    ! Start

    binTimeRange(1)%lower%type = INCLUSIVE
    binTimeRange(1)%upper%type = INCLUSIVE
    binTimeRange(1)%lower%value = binStartTime
    binTimeRange(1)%upper%value = binStopTime

    timeRanges%lower%type = INCLUSIVE
    timeRanges%upper%type = INCLUSIVE
    timeRanges%lower%value = intervals%lower
    timeRanges%upper%value = intervals%upper

!    binTimeRanges = andRangesPair(binTimeRange, timeRanges)
call andRanges(binTimeRange, timeRanges, binTimeRanges)
    allocate(binIntervals(size(binTimeRanges)))
    binIntervals%lower = binTimeRanges%lower%value
    binIntervals%upper  = binTimeRanges%upper%value

    deallocate(binTimeRanges)

  end function getBinIntervals


  !---------------------------------------------------------------------
  subroutine andIntervalsScalVec(inInterval, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two sets of intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: inInterval
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    type(IntervalT), pointer :: tempIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call andIntervalsVecVecVec((/inInterval/), andedIntervals, tempIntervals)
    deallocate(andedIntervals)
    andedIntervals => tempIntervals

  end subroutine andIntervalsScalVec

  !---------------------------------------------------------------------
  subroutine andIntervalsScalScalVec(intervalA, intervalB, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: intervalA, intervalB
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call andIntervalsVecVecVec((/intervalA/), (/intervalB/), andedIntervals)

  end subroutine andIntervalsScalScalVec

  !---------------------------------------------------------------------
  subroutine andIntervalsScalVecVec(intervalA, intervalsB, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two sets of intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: intervalA, intervalsB(:)
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call andIntervalsVecVecVec((/intervalA/), intervalsB, andedIntervals)

  end subroutine andIntervalsScalVecVec

  !---------------------------------------------------------------------
  subroutine andIntervalsVecScalVec(intervalsA, intervalB, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two sets of intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: intervalsA(:), intervalB
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call andIntervalsVecVecVec(intervalsA, (/intervalB/), andedIntervals)

  end subroutine andIntervalsVecScalVec

  !---------------------------------------------------------------------
  subroutine andIntervalsVecVec(inIntervals, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two sets of intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: inIntervals(:)
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    type(IntervalT), pointer :: tempIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call andIntervalsVecVecVec(inIntervals, andedIntervals, tempIntervals)
    deallocate(andedIntervals)
    andedIntervals => tempIntervals

  end subroutine andIntervalsVecVec

  !---------------------------------------------------------------------
  subroutine andIntervalsVecVecVec(intervalsA, intervalsB, andedIntervals)
  !---------------------------------------------------------------------
  ! ANDs the two sets of intervals together and leaves the result in
  ! andedIntervals.
  !
  ! Note: The returned pointer andedIntervals should be deallocated after
  ! use.
  !
  ! NOTE that no checking is done to ensure that the intervals are in order.
  !---------------------------------------------------------------------
!*** with doChecks a la andRanges??

    implicit none

    type(IntervalT), intent(in) :: intervalsA(:), intervalsB(:)
    type(IntervalT), pointer    :: andedIntervals(:) ! typedef in caltypes

    type(RangeT), pointer :: andedRanges(:)
    type(RangeT), allocatable :: rangesA(:), rangesB(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(rangesA(size(intervalsA)), rangesB(size(intervalsB)))

    rangesA%lower%type = INCLUSIVE
    rangesA%upper%type = EXCLUSIVE
    rangesA%lower%value = intervalsA%lower
    rangesA%upper%value = intervalsA%upper

    rangesB%lower%type = INCLUSIVE
    rangesB%upper%type = EXCLUSIVE
    rangesB%lower%value = intervalsB%lower
    rangesB%upper%value = intervalsB%upper

    call andRanges(rangesA, rangesB, andedRanges) ! in ssclib/dss_ranges_aux
    allocate(andedIntervals(size(andedRanges)))
    andedIntervals%lower = andedRanges%lower%value
    andedIntervals%upper = andedRanges%upper%value
    deallocate(andedRanges)

  end subroutine andIntervalsVecVecVec

  !---------------------------------------------------------------------
  subroutine andIntervalsScalScal(inInterval, andedInterval)
  !---------------------------------------------------------------------
  ! ANDs the two intervals together and leaves the result in
  ! andedInterval.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in)    :: inInterval
    type(IntervalT), intent(inout) :: andedInterval ! typedef in caltypes

    type(RangeT), pointer :: andedRange
    type(RangeT) :: rangeA, rangeB

    !-------------------------------------------------------------------
    ! Start

    rangeA%lower%type  = INCLUSIVE
    rangeA%upper%type  = EXCLUSIVE
    rangeA%lower%value = inInterval%lower
    rangeA%upper%value = inInterval%upper

    rangeB%lower%type  = INCLUSIVE
    rangeB%upper%type  = EXCLUSIVE
    rangeB%lower%value = andedInterval%lower
    rangeB%upper%value = andedInterval%upper

    call andRanges(rangeA, rangeB, andedRange)

    andedInterval%lower = andedRange%lower%value
    andedInterval%upper = andedRange%upper%value

  end subroutine andIntervalsScalScal

  !---------------------------------------------------------------------
  subroutine andIntervalsScalScalScal(intervalA, intervalB, andedInterval)
  !---------------------------------------------------------------------
  ! ANDs the two intervals together and leaves the result in
  ! andedInterval.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in)  :: intervalA, intervalB
    type(IntervalT), intent(out) :: andedInterval ! typedef in caltypes

    type(RangeT), pointer :: andedRange
    type(RangeT) :: rangeA, rangeB

    !-------------------------------------------------------------------
    ! Start

    rangeA%lower%type  = INCLUSIVE
    rangeA%upper%type  = EXCLUSIVE
    rangeA%lower%value = intervalA%lower
    rangeA%upper%value = intervalA%upper

    rangeB%lower%type  = INCLUSIVE
    rangeB%upper%type  = EXCLUSIVE
    rangeB%lower%value = intervalB%lower
    rangeB%upper%value = intervalB%upper

    call andRanges(rangeA, rangeB, andedRange)

    andedInterval%lower = andedRange%lower%value
    andedInterval%upper = andedRange%upper%value

  end subroutine andIntervalsScalScalScal


  !---------------------------------------------------------------------
  subroutine orIntervalsScalVec(inInterval, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: inInterval
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    type(IntervalT), pointer :: tempIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call orIntervalsVecVecVec((/inInterval/), oredIntervals, tempIntervals)
    deallocate(oredIntervals)
    oredIntervals => tempIntervals

  end subroutine orIntervalsScalVec

  !---------------------------------------------------------------------
  subroutine orIntervalsScalScalVec(intervalA, intervalB, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: intervalA, intervalB
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call orIntervalsVecVecVec((/intervalA/), (/intervalB/), oredIntervals)

  end subroutine orIntervalsScalScalVec

  !---------------------------------------------------------------------
  subroutine orIntervalsScalVecVec(intervalA, intervalsB, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: intervalA, intervalsB(:)
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call orIntervalsVecVecVec((/intervalA/), intervalsB, oredIntervals)

  end subroutine orIntervalsScalVecVec

  !---------------------------------------------------------------------
  subroutine orIntervalsVecScalVec(intervalsA, intervalB, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: intervalsA(:), intervalB
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call orIntervalsVecVecVec(intervalsA, (/intervalB/), oredIntervals)

  end subroutine orIntervalsVecScalVec

  !---------------------------------------------------------------------
  subroutine orIntervalsVecVec(inIntervals, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: inIntervals(:)
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    type(IntervalT), pointer :: tempIntervals(:) ! typedef in caltypes

    !-------------------------------------------------------------------
    ! Start

    call orIntervalsVecVecVec(inIntervals, oredIntervals, tempIntervals)
    deallocate(oredIntervals)
    oredIntervals => tempIntervals

  end subroutine orIntervalsVecVec

  !---------------------------------------------------------------------
  subroutine orIntervalsVecVecVec(intervalsA, intervalsB, oredIntervals)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !
  ! Note: The returned pointer oredIntervals should be deallocated after
  ! use.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in) :: intervalsA(:), intervalsB(:)
    type(IntervalT), pointer    :: oredIntervals(:) ! typedef in caltypes

    type(RangeT), pointer :: oredRanges(:)
    type(RangeT), allocatable :: rangesA(:), rangesB(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(rangesA(size(intervalsA)), rangesB(size(intervalsB)))

    rangesA%lower%type = INCLUSIVE
    rangesA%upper%type = EXCLUSIVE
    rangesA%lower%value = intervalsA%lower
    rangesA%upper%value = intervalsA%upper

    rangesB%lower%type = INCLUSIVE
    rangesB%upper%type = EXCLUSIVE
    rangesB%lower%value = intervalsB%lower
    rangesB%upper%value = intervalsB%upper

    call orRanges(rangesA, rangesB, oredRanges) ! in ssclib/dss_ranges_aux
    allocate(oredIntervals(size(oredRanges)))
    oredIntervals%lower = oredRanges%lower%value
    oredIntervals%upper = oredRanges%upper%value
    deallocate(oredRanges)

  end subroutine orIntervalsVecVecVec

  !---------------------------------------------------------------------
  subroutine orIntervalsScalScal(inInterval, oredInterval)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in)    :: inInterval
    type(IntervalT), intent(inout) :: oredInterval ! typedef in caltypes

    type(RangeT), pointer :: oredRange
    type(RangeT) :: rangeA, rangeB

    !-------------------------------------------------------------------
    ! Start

    rangeA%lower%type  = INCLUSIVE
    rangeA%upper%type  = EXCLUSIVE
    rangeA%lower%value = inInterval%lower
    rangeA%upper%value = inInterval%upper

    rangeB%lower%type  = INCLUSIVE
    rangeB%upper%type  = EXCLUSIVE
    rangeB%lower%value = oredInterval%lower
    rangeB%upper%value = oredInterval%upper

    call orRanges(rangeA, rangeB, oredRange)

    oredInterval%lower = oredRange%lower%value
    oredInterval%upper = oredRange%upper%value

  end subroutine orIntervalsScalScal

  !---------------------------------------------------------------------
  subroutine orIntervalsScalScalScal(intervalA, intervalB, oredInterval)
  !---------------------------------------------------------------------
  ! ORs the two intervals together and leaves the result in
  ! oredInterval.
  !---------------------------------------------------------------------

    implicit none

    type(IntervalT), intent(in)  :: intervalA, intervalB
    type(IntervalT), intent(out) :: oredInterval ! typedef in caltypes

    type(RangeT), pointer :: oredRange
    type(RangeT) :: rangeA, rangeB

    !-------------------------------------------------------------------
    ! Start

    rangeA%lower%type  = INCLUSIVE
    rangeA%upper%type  = EXCLUSIVE
    rangeA%lower%value = intervalA%lower
    rangeA%upper%value = intervalA%upper

    rangeB%lower%type  = INCLUSIVE
    rangeB%upper%type  = EXCLUSIVE
    rangeB%lower%value = intervalB%lower
    rangeB%upper%value = intervalB%upper

    call orRanges(rangeA, rangeB, oredRange)

    oredInterval%lower = oredRange%lower%value
    oredInterval%upper = oredRange%upper%value

  end subroutine orIntervalsScalScalScal

  !---------------------------------------------------------------------
  logical(bool) function isWithinInterval(time, intervals, includeBoundary)
  !---------------------------------------------------------------------
  ! NOTE! This sub has no protection against out-of-order intervals.

    implicit none

    real(double),    intent(in)           :: time
    type(IntervalT), intent(in)           :: intervals(:)
    logical(bool),   intent(in), optional :: includeBoundary

    logical(bool) :: localIncludeBoundary
integer :: maxLocVec(1), minLocVec(1)
    !-------------------------------------------------------------------
    ! Start

    if(present(includeBoundary)) then
      localIncludeBoundary = includeBoundary
    else
      localIncludeBoundary = .false.
    end if

    isWithinInterval = .false. ! default
  if (size(intervals) <= 0) return

    if (localIncludeBoundary) then
  if (minval(intervals%lower) >= time) return
  if (maxval(intervals%upper) <= time) return
      maxLocVec = maxloc(intervals%lower, (intervals%lower <= time))
      minLocVec = minloc(intervals%upper,  (intervals%upper  >= time))
      if (maxLocVec(1) == minLocVec(1)) isWithinInterval = .true.

    else
  if (minval(intervals%lower) > time) return
  if (maxval(intervals%upper) < time) return
      maxLocVec = maxloc(intervals%lower, (intervals%lower < time))
      minLocVec = minloc(intervals%upper,  (intervals%upper  > time))
      if (maxLocVec(1) == minLocVec(1)) isWithinInterval = .true.
    end if

  end function isWithinInterval

!  !---------------------------------------------------------------------
!  function checkIntervalOrder(intervals) result(status)
!  !---------------------------------------------------------------------
!
!!*** change name to intervalsAreWellFormed? and return just a boolean? Look at the analogous rangesAreWellFormed.
!
!    implicit none
!
!    type(IntervalT), intent(in) :: intervals(:)
!
!    integer :: status, numGtis
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    status = 0
!
!    numGtis = size(intervals)
!    if (numGtis <= 0) then
!      status = 1
!  return
!    end if
!
!    if (any(intervals%lower >= intervals%upper)) then
!      status = -1
!    else if (numGtis > 1) then
!      if (any(intervals(:numGtis-1)%upper >= intervals(2:numGtis)%lower)) then
!        status = -2
!      end if
!    end if
!
!  end function checkIntervalOrder
!
!  !---------------------------------------------------------------------
!  function intervalOrderType(intervals)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end function intervalOrderType

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module intervals_aux
