!+
! Module: splines
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains spline routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module splines

  use types

  use constants
  use linear
  use array_utils	! for largestTrueIndex().

  implicit none 

  private 

  public calcCubicSplines
  public cubicSplineInterpolate

contains 

  !---------------------------------------------------------------------
  subroutine calcCubicSplines(x, y, loBoundType, loBoundValue, hiBoundType&
  , hiBoundValue, y2, status)
  !---------------------------------------------------------------------
  ! This is designed to work in conjunction with subroutine
  ! cubicSplineInterpolate() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values
  ! plus corresponding y values, calculates the values at all x of the
  ! second derivatives (returned in y2) of the spline function which is
  ! continuous in its second derivative and which obeys the boundary
  ! conditions.
  !
  ! In terms of the variable names used, the following equation is solved:
  !
  !		triDiag * y2 = b.
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: x(:),&
                                           y(size(x)),&
                                           loBoundValue,&
                                           hiBoundValue
    character(*), intent(in)            :: loBoundType,&
                                           hiBoundType
    real(single), intent(out)           :: y2(size(x))
    integer,      intent(out), optional :: status

    integer :: localStatus, n, i
    real(single) :: triDiag(-1:1, size(x)), b(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    y2 = 0.0

    n = size(x)
  if (n <= 0) return ! no point in going further.

    select case(loBoundType)
    case('fixed1st') ! loBoundValue is taken to be the value of the
    ! first derivative at the lowest x sample.

      triDiag(0, 1) = (x(2) - x(1)) / 3.0
      triDiag(1, 1) = (x(2) - x(1)) / 6.0
      b(1) = (y(2) - y(1)) / (x(2) - x(1)) - loBoundValue
    case('fixed2nd') ! loBoundValue is taken to be the value of the
    ! second derivative at the lowest x sample.

      triDiag(0, 1) = 1.0
      triDiag(1, 1) = 0.0
      b(1) = loBoundValue

    case default
      if (present(status)) status = -1
  return
    end select

    select case(hiBoundType)
    case('fixed1st') ! hiBoundType is taken to be the value of the
    ! first derivative at the highest x sample.

      triDiag(-1, n) = (x(n) - x(n-1)) / 6.0
      triDiag( 0, n) = (x(n) - x(n-1)) / 3.0
      b(n) = hiBoundValue - (y(n) - y(n-1)) / (x(n) - x(n-1))

    case('fixed2nd') ! hiBoundType is taken to be the value of the
    ! second derivative at the highest x sample.

      triDiag(-1, n) = 0.0
      triDiag( 0, n) = 1.0
      b(n) = hiBoundValue

    case default
      if (present(status)) status = -1
  return
    end select

    ! Fill up the middle rows of triDiag and b:

    do i = 2, n - 1
      triDiag(-1, i) = (x( i ) - x(i-1)) / 6.0
      triDiag( 0, i) = (x(i+1) - x(i-1)) / 3.0
      triDiag( 1, i) = (x(i+1) - x( i )) / 6.0

      b(i) = (y(i+1) - y( i )) / (x(i+1) - x( i ))&
           - (y( i ) - y(i-1)) / (x( i ) - x(i-1))
    end do

    triDiag(-1, 1) = 0.0 ! always.
    triDiag( 1, n) = 0.0 ! always.

    ! Now solve the system of equations:

    call solveLinearTriDiag(triDiag, b, y2)

    if (present(status)) status = localStatus

  end subroutine calcCubicSplines

  !---------------------------------------------------------------------
  subroutine cubicSplineInterpolate(x, y, y2, xIn, yOut, status)
  !---------------------------------------------------------------------
  ! This is designed to work in conjunction with subroutine
  ! calcCubicSplines() to provide a cubic-spline-type interpolation.
  ! The present subroutine, given an ordered series of unique x values,
  ! plus the corresponding y values, plus the corresponding values of
  ! the second derivative y2, and finally plus a single x value xIn,
  ! performs the interpolation using the formula
  !
  !	y = a*y(i) + b*y(i+1) + c*y2(i) + d*y2(i+1)
  !
  ! where
  !
  !	     x(i+1) - xIn
  !	a = ---------------, b = 1 - a,
  !	     x(i+1) - x(i)
  !
  !	     1
  !	c = --- a*(a*a - 1)*(x(i+1) - x(i))^2
  !	     6
  ! and
  !	     1
  !	d = --- b*(b*b - 1)*(x(i+1) - x(i))^2
  !	     6
  !
  ! to return in the variable yOut the value of y at xIn.
  !
  ! Note that xIn must be >= x(1) and <= x(n).
  !
  ! This routine was written with the help of Numerical Recipes for Fortran.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: x(:), y(size(x)), y2(size(x)), xIn
    real(single), intent(out)           :: yOut
    integer,      intent(out), optional :: status

    integer :: localStatus, n, i
    real(single) :: a, b, denom, commonTerm

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    yOut = 0.0

    n = size(x)

    if (xIn < x(1) .or. xIn > x(n)) then
      if (present(status)) status = -1
  return
    end if

    i = largestTrueIndex(logical(x <= xIn, bool)) ! in array_utils
    if (i == n) i = n - 1

    denom = x(i+1) - x(i)
    a = (x(i+1) - xIn)  / denom
    b = (xIn    - x(i)) / denom

    commonTerm = denom * denom / 6.0
    yOut = a * (y(i)   + (a * a - 1.0) * commonTerm * y2(i))&
         + b * (y(i+1) + (b * b - 1.0) * commonTerm * y2(i+1))

    if (present(status)) status = localStatus

  end subroutine cubicSplineInterpolate

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module splines
