!
! Module: wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Routines to facilitate handling of wcs values.
! 
! Comments: TODO:
!	As of 2004-01-30 the library generally supports the WCS system outlined in Greisen and Calabretta A&A 395, 1061-1075 (2002). However, a couple of caveats: (i) only the RA---TAN/DEC--TAN projection is supported; (ii) if CD1_1 is present, all CDi_j kwds are required to be present; (iii) PCi_j kwds are not yet supported.
!
! $Id: wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module wcs_aux

  use dal
  use errorHandling	! used in dumpWcs and the pixels<->tan routines.

  use constants
  use dal_aux	! contains subs getRealKwd(), getRealKwd(), getStringKwd().
  use ssc_misc	! for approxEqual
  use array_utils	! for vectorCross

  implicit none 

  private 

  public initializeWcsT
  public copyWcsT
  private checkWcsType

  public hasWcs
  public readWcs
  public writeWcs
  public deleteWcs
  public dumpWcs

  public easifyWcsType
  public wcsAreEquivalent
  public wcsPixelAreaWorld
  public checkImageEquivalence
  public convertOldToNew
  public convertNewToOld

  public pixelsToTan
  public tanToPixels
  public tanRadiusToPixelsEllipse

public readWcsSourceList
interface readWcsSourceList
  module procedure readWcsSourceListSet, readWcsSourceListName
end interface
!*** this is necessary until src lists conform to ogip.


  interface hasWcs
    module procedure hasWcsColumnsName, hasWcsColumns
    module procedure hasWcsArray
    module procedure hasWcsSet, hasWcsSetName
    module procedure hasWcsTable
  end interface

  interface readWcs
    module procedure readWcsColumnsName, readWcsColumns
    module procedure readWcsSetArray, readWcsArray
    module procedure readWcsSet, readWcsSetName, readWcsSetArrayName
    module procedure readWcsTable, readWcsSetArrayNumber
  end interface

  interface writeWcs
    module procedure writeWcsColumns, writeWcsArray
  end interface

  interface easifyWcsType
    module procedure easifyWcsTypeScalar
    module procedure easifyWcsTypeVector
  end interface

  interface tanToPixels
    module procedure tanToPixelsScalar, tanToPixelsVector, tanToPixelsArray
  end interface

  interface tanRadiusToPixelsEllipse
    module procedure tanRadToPixEllipseScalar, tanRadToPixEllipseVector
!, tanRadToPixEllipseArray
  end interface

  interface pixelsToTan
    module procedure pixelsToTanScalar, pixelsToTanVector, pixelsToTanArray
  end interface

  type, public :: WcsAxisT
    character(len=8) ::&
      ctype,&		! CTYPE
      worldUnit		! 
    real(double) ::&
      matrixOnAxis,&	! CD1_
      matrixOffAxis,&	! CD2_
      refPixels,&	! CRPIX
      refWorld		! CRVAL
  end type WcsAxisT

  type, public :: WcsT
    type(WcsAxisT) :: x, y
    logical(bool) ::&
      crotFound
  end type WcsT

logical :: useNewWcsStyle = .false.

contains 

  !---------------------------------------------------------------------
  logical function hasWcsColumnsName(tab, xColName, yColName)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: xColName, yColName

    type(ColumnT) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(tab, xColName)
    yCol = column(tab, yColName)

    hasWcsColumnsName = hasWcsColumns(xCol, yCol)

    call release(xCol)
    call release(yCol)

  end function hasWcsColumnsName

  !---------------------------------------------------------------------
  logical function hasWcsColumns(xCol, yCol)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    hasWcsColumns = .false. ! default

    if (    hasAttribute(xCol, 'TCRPX')&
    .and.   hasAttribute(yCol, 'TCRPX')&
    .and.   hasAttribute(xCol, 'TCRVL')&
    .and.   hasAttribute(yCol, 'TCRVL')&
    .and.   hasAttribute(xCol, 'TCTYP')&
    .and.   hasAttribute(yCol, 'TCTYP')&
    .and. ((hasAttribute(xCol, 'TCDLT')&
    .and.   hasAttribute(yCol, 'TCDLT'))&
    .or.   (hasAttribute(xCol, 'TCD1_')&
    .and.   hasAttribute(yCol, 'TCD1_')&
    .and.   hasAttribute(xCol, 'TCD2_')&
    .and.   hasAttribute(yCol, 'TCD2_'))))&
      hasWcsColumns = .true.

  end function hasWcsColumns

  !---------------------------------------------------------------------
  logical function hasWcsSetName(setName)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    hasWcsSetName = .false. ! default

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    if (hasWcsArray(inArray)) then
      hasWcsSetName = .true.
      if (hasWcsSet(set)) call warning('setAlsoHasWcs',"ssclib:hasWcsSetName: &
        &Both the dataset header and the array header contain WCS keywords. &
        &The correct place for them is in the array header only.")

    else if (hasWcsSet(set)) then
      hasWcsSetName = .true.
      call warning('onlySetHadWcs',"ssclib:hasWcsSetName: The dataset header &
        &contains WCS keywords but not the array header. It should be the other &
        &way around.")
    end if

    call release(set)

  end function hasWcsSetName

  !---------------------------------------------------------------------
  logical function hasWcsSet(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    !-------------------------------------------------------------------
    ! Start

    hasWcsSet = .false. ! default

    if (    hasAttribute(set, 'CRPIX1')&
    .and.   hasAttribute(set, 'CRPIX2')&
    .and.   hasAttribute(set, 'CRVAL1')&
    .and.   hasAttribute(set, 'CRVAL2')&
    .and.   hasAttribute(set, 'CTYPE1')&
    .and.   hasAttribute(set, 'CTYPE2')&
    .and. ((hasAttribute(set, 'CDELT1')&
    .and.   hasAttribute(set, 'CDELT2'))&
    .or.   (hasAttribute(set, 'CD1_1')&
    .and.   hasAttribute(set, 'CD1_2')&
    .and.   hasAttribute(set, 'CD2_1')&
    .and.   hasAttribute(set, 'CD2_2'))))&
      hasWcsSet = .true.

  end function hasWcsSet

  !---------------------------------------------------------------------
  logical function hasWcsArray(inArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray

    !-------------------------------------------------------------------
    ! Start

    hasWcsArray = .false. ! default

    if (    hasAttribute(inArray, 'CRPIX1')&
    .and.   hasAttribute(inArray, 'CRPIX2')&
    .and.   hasAttribute(inArray, 'CRVAL1')&
    .and.   hasAttribute(inArray, 'CRVAL2')&
    .and.   hasAttribute(inArray, 'CTYPE1')&
    .and.   hasAttribute(inArray, 'CTYPE2')&
    .and. ((hasAttribute(inArray, 'CDELT1')&
    .and.   hasAttribute(inArray, 'CDELT2'))&
    .or.   (hasAttribute(inArray, 'CD1_1')&
    .and.   hasAttribute(inArray, 'CD1_2')&
    .and.   hasAttribute(inArray, 'CD2_1')&
    .and.   hasAttribute(inArray, 'CD2_2'))))&
      hasWcsArray = .true.

  end function hasWcsArray

  !---------------------------------------------------------------------
  logical function hasWcsTable(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab

    !-------------------------------------------------------------------
    ! Start

    hasWcsTable = .false. ! default

    if (    hasAttribute(tab, 'CRPIX1')&
    .and.   hasAttribute(tab, 'CRPIX2')&
    .and.   hasAttribute(tab, 'CRVAL1')&
    .and.   hasAttribute(tab, 'CRVAL2')&
    .and.   hasAttribute(tab, 'CTYPE1')&
    .and.   hasAttribute(tab, 'CTYPE2')&
    .and. ((hasAttribute(tab, 'CDELT1')&
    .and.   hasAttribute(tab, 'CDELT2'))&
    .or.   (hasAttribute(tab, 'CD1_1')&
    .and.   hasAttribute(tab, 'CD1_2')&
    .and.   hasAttribute(tab, 'CD2_1')&
    .and.   hasAttribute(tab, 'CD2_2'))))&
      hasWcsTable = .true.

  end function hasWcsTable

  !---------------------------------------------------------------------
  subroutine initializeWcsT(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start

    wcs%x%ctype          = ''
    wcs%x%worldUnit      = ''
    wcs%x%matrixOnAxis   = 1.0d0
    wcs%x%matrixOffAxis  = 0.0d0
    wcs%x%refPixels      = 0.0d0
    wcs%x%refWorld       = 0.0d0

    wcs%y%ctype          = ''
    wcs%y%worldUnit      = ''
    wcs%y%matrixOnAxis   = 1.0d0
    wcs%y%matrixOffAxis  = 0.0d0
    wcs%y%refPixels      = 0.0d0
    wcs%y%refWorld       = 0.0d0

    wcs%crotFound        = .false.

  end subroutine initializeWcsT

  !---------------------------------------------------------------------
  function copyWcsT(inWcs) result(outWcs)
  !---------------------------------------------------------------------
!****** don't need this.

    implicit none

    type(WcsT), intent(in) :: inWcs
    type(WcsT)             :: outWcs

    !-------------------------------------------------------------------
    ! Start

    outWcs%x%ctype          = inWcs%x%ctype
    outWcs%x%worldUnit      = inWcs%x%worldUnit
    outWcs%x%matrixOnAxis   = inWcs%x%matrixOnAxis
    outWcs%x%matrixOffAxis  = inWcs%x%matrixOffAxis
    outWcs%x%refPixels      = inWcs%x%refPixels
    outWcs%x%refWorld       = inWcs%x%refWorld

    outWcs%y%ctype          = inWcs%y%ctype
    outWcs%y%worldUnit      = inWcs%y%worldUnit
    outWcs%y%matrixOnAxis   = inWcs%y%matrixOnAxis
    outWcs%y%matrixOffAxis  = inWcs%y%matrixOffAxis
    outWcs%y%refPixels      = inWcs%y%refPixels
    outWcs%y%refWorld       = inWcs%y%refWorld

    outWcs%crotFound      = inWcs%crotFound

  end function copyWcsT

  !---------------------------------------------------------------------
  subroutine checkWcsType(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    select case(wcs%x%ctype)
    case('RA---TAN')
    case default
call error('notSupported',"ssclib:checkWcsType: This value of CTYPE1 is not yet supported.")
    end select

    select case(wcs%y%ctype)
    case('DEC--TAN')
    case default
call error('notSupported',"ssclib:checkWcsType: This value of CTYPE2 is not yet supported.")
    end select

  end subroutine checkWcsType

  !---------------------------------------------------------------------
  subroutine getDefaultXWorldUnit(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    select case(wcs%x%ctype)
    case('RA---TAN')
      wcs%x%worldUnit = 'deg'
    case default
call error('notSupported',"ssclib:getDefaultXWorldUnit: This value of CTYPE1 is not yet supported.")
    end select

  end subroutine getDefaultXWorldUnit

  !---------------------------------------------------------------------
  subroutine getDefaultYWorldUnit(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    select case(wcs%y%ctype)
    case('DEC--TAN')
      wcs%y%worldUnit = 'deg'
    case default
call error('notSupported',"ssclib:getDefaultYWorldUnit: This value of CTYPE2 is not yet supported.")
    end select

  end subroutine getDefaultYWorldUnit

  !---------------------------------------------------------------------
  subroutine readWcsColumnsName(tab, xColName, yColName, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: xColName, yColName
    type(WcsT),   intent(out) :: wcs

    type(ColumnT) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(tab, xColName)
    yCol = column(tab, yColName)

    call readWcsColumns(xCol, yCol, wcs)

    call release(xCol)
    call release(yCol)

  end subroutine readWcsColumnsName

  !---------------------------------------------------------------------
  subroutine readWcsColumns(xCol, yCol, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype          = stringAttribute(xCol, 'TCTYP')
    wcs%y%ctype          = stringAttribute(yCol, 'TCTYP')

    wcs%x%refPixels      = real64Attribute(xCol, 'TCRPX')
    wcs%y%refPixels      = real64Attribute(yCol, 'TCRPX')
    wcs%x%refWorld       = real64Attribute(xCol, 'TCRVL')
    wcs%y%refWorld       = real64Attribute(yCol, 'TCRVL')

    wcs%crotFound = .false. ! default
    if (hasAttribute(xCol, "TCD1_")) then
      wcs%x%matrixOnAxis  = real64Attribute(xCol, 'TCD1_')
      wcs%x%matrixOffAxis = real64Attribute(xCol, 'TCD2_')
      wcs%y%matrixOffAxis = real64Attribute(yCol, 'TCD1_')
      wcs%y%matrixOnAxis  = real64Attribute(yCol, 'TCD2_')

    else
      xPixelSizeWorld = real64Attribute(xCol, 'TCDLT')
      yPixelSizeWorld = real64Attribute(yCol, 'TCDLT')
      if (hasAttribute(yCol, "TCROT")) then
        rotAngle = real64Attribute(yCol, 'TCROT')
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0

! See comments in writeWcsColumns.
if (hasAttribute(parent(xCol), 'TMPTCROT')) then
  call warning('tcrotSubstituteFound',"ssclib:readWcsColumns: The DAL &
    &interface does not yet enable one to write the WCS roll angle keyword &
    &as a column attribute; until this is fixed, I'm writing this value to &
    &the TMPTCROT keyword in the table. Such a keyword has been detected and &
    &the roll angle will be read from it.")
  rotAngle = real64Attribute(parent(xCol), 'TMPTCROT') * PiOn180
end if
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(xCol, "TCUNI")) then
      wcs%x%worldUnit = stringAttribute(xCol, 'TCUNI')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(yCol, "TCUNI")) then
      wcs%y%worldUnit = stringAttribute(yCol, 'TCUNI')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsColumns

  !---------------------------------------------------------------------
  subroutine readWcsSetName(setName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName
    type(WcsT),   intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetName

  !---------------------------------------------------------------------
  subroutine readWcsSetArrayName(setName, arrayName, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName, arrayName
    type(WcsT),   intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetArrayName

  !---------------------------------------------------------------------
  subroutine readWcsSetArrayNumber(setName, arrayNumber, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),   intent(in)  :: setName
    integer(int32), intent(in)  :: arrayNumber
    type(WcsT),     intent(out) :: wcs

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayNumber, READ)
    call readWcsSetArray(set, inArray, wcs)

    call release(set)

  end subroutine readWcsSetArrayNumber

  !---------------------------------------------------------------------
  subroutine readWcsSetArray(set, inArray, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    type(DataSetT), intent(in)  :: set
    type(ArrayT),   intent(in)  :: inArray
    type(WcsT),     intent(out) :: wcs

    type(WcsT) :: setWcs

    !-------------------------------------------------------------------
    ! Start:

    if (hasWcsSet(set)) then ! set contains the wcs keywords.
      call readWcsSet(set, setWcs)
      ! Now check to see if the array block also has a set:

      if (hasWcsArray(inArray)) then ! array also contains the wcs keywords.
        if (hasWcsSet(set)) call warning('setAlsoHasWcs',"ssclib:readWcsSetArray: &
          &Both the dataset header and the array header contain WCS keywords. &
          &The correct place for them is in the array header only.")
        call readWcsArray(inArray, wcs)

      else ! array doesn't contains any wcs keywords.
        call warning('onlySetHadWcs',"ssclib:readWcsSetArray: The dataset header &
          &contains WCS keywords but not the array header. It should be the other &
          &way around.")
        wcs = setWcs
      end if
    else ! set header doesn't contain any wcs keywords. Check the array:
      call readWcsArray(inArray, wcs)
    end if

  end subroutine readWcsSetArray

  !---------------------------------------------------------------------
  subroutine readWcsArray(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in)  :: inArray
    type(WcsT),   intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(inArray)
    call readWcsGeneric(generic, wcs)

  end subroutine readWcsArray

  !---------------------------------------------------------------------
  subroutine readWcsSet(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsT),     intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(set)
    call readWcsGeneric(generic, wcs)

  end subroutine readWcsSet

  !---------------------------------------------------------------------
  subroutine readWcsTable(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    type(WcsT),   intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(tab)
    call readWcsGeneric(generic, wcs)

  end subroutine readWcsTable

  !---------------------------------------------------------------------
  subroutine readWcsGeneric(generic, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(AttributableT), intent(in)  :: generic
    type(WcsT),          intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(generic, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(generic, 'CTYPE2')

    wcs%x%refPixels = real64Attribute(generic, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(generic, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(generic, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(generic, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(generic, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(generic, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(generic, 'CD2_1')
      wcs%y%matrixOffAxis = real64Attribute(generic, 'CD1_2')
      wcs%y%matrixOnAxis  = real64Attribute(generic, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(generic, 'CDELT1')
      yPixelSizeWorld = real64Attribute(generic, 'CDELT2')
      if (hasAttribute(generic, "CROTA2")) then
        rotAngle = real64Attribute(generic, 'CROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(generic, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(generic, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(generic, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(generic, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsGeneric


  !---------------------------------------------------------------------
  subroutine readWcsSetArray_20040818(set, inArray, wcs)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    type(DataSetT), intent(in)  :: set
    type(ArrayT),   intent(in)  :: inArray
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld
    integer :: status

    !-------------------------------------------------------------------
    ! Start:

    call getStringKwd("CTYPE1", set, inArray, wcs%x%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%y%ctype)

!    call checkWcsType(wcs)

    call getRealKwd("CRPIX1",   set, inArray, wcs%x%refPixels)
    call getRealKwd("CRPIX2",   set, inArray, wcs%y%refPixels)
    call getRealKwd("CRVAL1",   set, inArray, wcs%x%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%y%refWorld)

    wcs%crotFound = .false. ! default
    call getRealKwd("CD1_1", set, inArray, wcs%x%matrixOnAxis, status)
    if (status == 0) then ! the keyword was found, get the others:
      call getRealKwd("CD2_1", set, inArray, wcs%x%matrixOffAxis)
!      call getRealKwd("CD1_2", set, inArray, wcs%y%matrixOnAxis)
!      call getRealKwd("CD2_2", set, inArray, wcs%y%matrixOffAxis)
call getRealKwd("CD1_2", set, inArray, wcs%y%matrixOffAxis)
call getRealKwd("CD2_2", set, inArray, wcs%y%matrixOnAxis)

    else
      call getRealKwd("CDELT1", set, inArray, xPixelSizeWorld)
      call getRealKwd("CDELT2", set, inArray, yPixelSizeWorld)

      call getRealKwd("CROTA2", set, inArray, rotAngle, status)
      if (status == 0) then ! the CROTA keyword was found:
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    call getStringKwd("CUNIT1", set, inArray, wcs%x%worldUnit, status)
    if (status .ne. 0) then
      call getDefaultXWorldUnit(wcs)
    end if

    call getStringKwd("CUNIT2", set, inArray, wcs%y%worldUnit, status)
    if (status .ne. 0) then
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSetArray_20040818

  !---------------------------------------------------------------------
  subroutine readWcsArray_20040818(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in)  :: inArray
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(inArray, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(inArray, 'CTYPE2')

!    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(inArray, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(inArray, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(inArray, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(inArray, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(inArray, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(inArray, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(inArray, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(inArray, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(inArray, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(inArray, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(inArray, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(inArray, 'CDELT1')
      yPixelSizeWorld = real64Attribute(inArray, 'CDELT2')
      if (hasAttribute(inArray, "CROTA2")) then
        rotAngle = real64Attribute(inArray, 'CROTA2')
        wcs%crotFound = .true.

      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(inArray, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(inArray, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(inArray, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(inArray, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsArray_20040818

  !---------------------------------------------------------------------
  subroutine readWcsSet_20040818(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(set, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(set, 'CTYPE2')

!    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(set, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(set, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(set, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(set, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(set, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(set, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(set, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(set, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(set, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(set, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(set, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(set, 'CDELT1')
      yPixelSizeWorld = real64Attribute(set, 'CDELT2')
      if (hasAttribute(set, "CROTA2")) then
        rotAngle = real64Attribute(set, 'CROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(set, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(set, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(set, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(set, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSet_20040818

  !---------------------------------------------------------------------
  subroutine readWcsTable_20040818(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    type(WcsT),   intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(tab, 'CTYPE1')
    wcs%y%ctype     = stringAttribute(tab, 'CTYPE2')

!    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(tab, 'CRPIX1')
    wcs%y%refPixels = real64Attribute(tab, 'CRPIX2')
    wcs%x%refWorld  = real64Attribute(tab, 'CRVAL1')
    wcs%y%refWorld  = real64Attribute(tab, 'CRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(tab, "CD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(tab, 'CD1_1')
      wcs%x%matrixOffAxis = real64Attribute(tab, 'CD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(tab, 'CD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(tab, 'CD2_2')
wcs%y%matrixOffAxis = real64Attribute(tab, 'CD1_2')
wcs%y%matrixOnAxis  = real64Attribute(tab, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(tab, 'CDELT1')
      yPixelSizeWorld = real64Attribute(tab, 'CDELT2')
      if (hasAttribute(tab, "CROTA2")) then
        rotAngle = real64Attribute(tab, 'CROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(tab, "CUNIT1")) then
      wcs%x%worldUnit = stringAttribute(tab, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(tab, "CUNIT2")) then
      wcs%y%worldUnit = stringAttribute(tab, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsTable_20040818

  !---------------------------------------------------------------------
  subroutine readWcsSourceListSet(set, wcs)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsT),     intent(out) :: wcs

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%ctype     = stringAttribute(set, 'TCTYPE1')
    wcs%y%ctype     = stringAttribute(set, 'TCTYPE2')

!    call checkWcsType(wcs)

    wcs%x%refPixels = real64Attribute(set, 'TCRPIX1')
    wcs%y%refPixels = real64Attribute(set, 'TCRPIX2')
    wcs%x%refWorld  = real64Attribute(set, 'TCRVAL1')
    wcs%y%refWorld  = real64Attribute(set, 'TCRVAL2')

    wcs%crotFound = .false. ! default
    if (hasAttribute(set, "TCD1_1")) then
      wcs%x%matrixOnAxis  = real64Attribute(set, 'TCD1_1')
      wcs%x%matrixOffAxis = real64Attribute(set, 'TCD2_1')
!      wcs%y%matrixOnAxis  = real64Attribute(set, 'TCD1_2')
!      wcs%y%matrixOffAxis = real64Attribute(set, 'TCD2_2')
wcs%y%matrixOffAxis = real64Attribute(set, 'TCD1_2')
wcs%y%matrixOnAxis  = real64Attribute(set, 'TCD2_2')

    else
      xPixelSizeWorld = real64Attribute(set, 'TCDELT1')
      yPixelSizeWorld = real64Attribute(set, 'TCDELT2')
      if (hasAttribute(set, "TCROTA2")) then
        rotAngle = real64Attribute(set, 'TCROTA2')
        wcs%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(set, "TCUNI1")) then
      wcs%x%worldUnit = stringAttribute(set, 'TCUNI1')
    else
      call getDefaultXWorldUnit(wcs)
    end if

    if (hasAttribute(set, "TCUNI2")) then
      wcs%y%worldUnit = stringAttribute(set, 'TCUNI2')
    else
      call getDefaultYWorldUnit(wcs)
    end if

  end subroutine readWcsSourceListSet

  !---------------------------------------------------------------------
  subroutine readWcsSourceListName(setName, wcs)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    character(*), intent(in) :: setName
    type(WcsT), intent(out) :: wcs

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    call readWcsSourceListSet(set, wcs)
    call release(set)

  end subroutine readWcsSourceListName

  !---------------------------------------------------------------------
  subroutine writeWcsColumns(xCol, yCol, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsT),    intent(in) :: wcs

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(xCol, "TCTYP", wcs%X%ctype)
    call setAttribute(xCol, "TCRPX", wcs%X%refPixels)
    call setAttribute(xCol, "TCRVL", wcs%X%refWorld)
    call setAttribute(xCol, 'TCUNI', wcs%X%worldUnit)

    call setAttribute(yCol, "TCTYP", wcs%Y%ctype)
    call setAttribute(yCol, "TCRPX", wcs%Y%refPixels)
    call setAttribute(yCol, "TCRVL", wcs%Y%refWorld)
    call setAttribute(yCol, 'TCUNI', wcs%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(xCol, 'TCD1_',  wcs%X%matrixOnAxis)
      call setAttribute(xCol, 'TCD2_',  wcs%X%matrixOffAxis)

!      call setAttribute(yCol, 'TCD1_',  wcs%Y%matrixOnAxis)
!      call setAttribute(yCol, 'TCD2_',  wcs%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD1_',  wcs%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD2_',  wcs%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(xCol, "TCDLT", xPixelSizeWorld)
      call setAttribute(yCol, "TCDLT", yPixelSizeWorld)
!!!      call setAttribute(yCol, "TCROT", rotAngle)

!*** apparently the dal doesn't like the TCROT keyword. I've removed it for the time being to get rid of the dal warning which always used to occur when writeWcsColumns was invoked. However I've SPRed the dal (SOC SPR 2523) and if this gets accepted you can decomment the setAttribute line and get rid of this workaround.

if (rotAngle .ne. 0.0) then
  call warning('unableToWriteTcrot',"ssclib:writeWcsColumns: The DAL &
    &interface does not yet enable one to write the WCS roll angle keyword &
    &as a column attribute. I'm writing this value to the TMPTCROT keyword &
    &in the table.")

  call setAttribute(parent(xCol), 'TMPTCROT', rotAngle/PiOn180, 'deg'&
    , 'Temporary storage of column-wcs roll angle.')

end if
    end if

  end subroutine writeWcsColumns

  !---------------------------------------------------------------------
  subroutine writeWcsArray(inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(WcsT),   intent(in) :: wcs

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(inArray, 'CRPIX1', wcs%X%refPixels)
    call setAttribute(inArray, 'CRPIX2', wcs%Y%refPixels)
    call setAttribute(inArray, 'CRVAL1', wcs%X%refWorld)
    call setAttribute(inArray, 'CRVAL2', wcs%Y%refWorld)
    call setAttribute(inArray, 'CTYPE1', wcs%X%ctype)
    call setAttribute(inArray, 'CTYPE2', wcs%Y%ctype)

    call setAttribute(inArray, 'CUNIT1', wcs%X%worldUnit)
    call setAttribute(inArray, 'CUNIT2', wcs%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(inArray, 'CD1_1',  wcs%X%matrixOnAxis)
      call setAttribute(inArray, 'CD2_1',  wcs%X%matrixOffAxis)
!      call setAttribute(inArray, 'CD1_2',  wcs%Y%matrixOnAxis)
!      call setAttribute(inArray, 'CD2_2',  wcs%Y%matrixOffAxis)
call setAttribute(inArray, 'CD1_2',  wcs%Y%matrixOffAxis)
call setAttribute(inArray, 'CD2_2',  wcs%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(inArray, 'CROTA2', rotAngle)
      call setAttribute(inArray, 'CDELT1', xPixelSizeWorld)
      call setAttribute(inArray, 'CDELT2', yPixelSizeWorld)
    end if

  end subroutine writeWcsArray

  !---------------------------------------------------------------------
  subroutine deleteWcs(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    integer, parameter :: numRootStrs = 6, maxNumDims = 3
    character(5) :: rootStrs(numRootStrs)
    integer :: i, j
    character(6) :: str1
    character(5) :: str2

    !-------------------------------------------------------------------
    ! Start

    rootStrs = (/'CRPIX', 'CRVAL', 'CTYPE', 'CDELT', 'CROTA', 'CUNIT'/)

    do i = 1, maxNumDims
      do j = 1, numRootStrs
        write(str1, '(a5, i1)') rootStrs(j), i
        if (hasAttribute(set, str1)) call deleteAttribute(set, str1)
      end do
      do j = 1, maxNumDims
        write(str2, '(a2, i1, a1, i1)') 'CD', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
        write(str2, '(a2, i1, a1, i1)') 'PC', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
      end do
    end do

  end subroutine deleteWcs

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeScalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeScalar

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeVector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeVector

  !---------------------------------------------------------------------
  subroutine dumpWcs(wcs, verbosityArg)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs
    integer(int32), intent(in), optional :: verbosityArg

    integer(int32) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArg)) then
      verbosity = verbosityArg
    else
      verbosity = VERBOSE
    end if

    call message(AppMsg, verbosity, 'WCS dump:')
    call message(AppMsg, verbosity, '  X axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%x%ctype)
    call message(AppMsg, verbosity, '    CUNIT: '//wcs%x%worldUnit)
    write(errstr, '(a, es23.16)')   '    CRVAL:', wcs%x%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CRPIX:', wcs%x%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD1_1:', wcs%x%matrixOnAxis
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD2_1:', wcs%x%matrixOffAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CDELT:', wcs%x%pixelSizeWorld
!    call message(AppMsg, verbosity)

    call message(AppMsg, verbosity, '  Y axis:')
    call message(AppMsg, verbosity, '    CTYPE: '//wcs%y%ctype)
    call message(AppMsg, verbosity, '    CUNIT: '//wcs%y%worldUnit)
    write(errstr, '(a, es23.16)')   '    CRVAL:', wcs%y%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CRPIX:', wcs%y%refPixels
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CD1_2:', wcs%y%matrixOnAxis
write(errstr, '(a, es23.16)')   '    CD1_2:', wcs%y%matrixOffAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CD2_2:', wcs%y%matrixOffAxis
write(errstr, '(a, es23.16)')   '    CD2_2:', wcs%y%matrixOnAxis
    call message(AppMsg, verbosity)
!    write(errstr, '(a, es23.16)')   '    CDELT:', wcs%y%pixelSizeWorld
!    call message(AppMsg, verbosity)
!**** fix this

  end subroutine dumpWcs

  !---------------------------------------------------------------------
  subroutine pixelsToTanScalar(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------
  ! See pixelsToTanVector for a description.

    implicit none

    real(single), intent(in)  :: xPixels, yPixels
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: tanX, tanY

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

!    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

    xPixelsVec(1) = xPixels
    yPixelsVec(1) = yPixels

    call pixelsToTanVector(xPixelsVec, yPixelsVec, wcs, tanXVec, tanYVec)

    tanX = tanXVec(1)
    tanY = tanYVec(1)

  end subroutine pixelsToTanScalar

  !---------------------------------------------------------------------
  subroutine pixelsToTanVector(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is implemented in the present subroutine, then the result x_i
  ! is converted to my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: xPixels(:), yPixels(size(xPixels))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: tanX(size(xPixels)), tanY(size(xPixels))

    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY
    real(double), allocatable ::&
      intermediateWorldX(:), intermediateWorldY(:)
    real(single), allocatable :: relPixelsX(:), relPixelsY(:)
    character(3) :: xProjTypeStr, yProjTypeStr

    !-------------------------------------------------------------------
    ! Start

    allocate(intermediateWorldX(size(xPixels)))
    allocate(intermediateWorldY(size(xPixels)))
    allocate(relPixelsX(size(xPixels)), relPixelsY(size(xPixels)))

!    call checkWcsType(wcs)
    xProjTypeStr = getProjectionType(wcs%x)
    yProjTypeStr = getProjectionType(wcs%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:pixelsToTanVector: Non-TAN &
&projections are not yet supported.")

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
&than 'deg' are not yet supported.")
    end select

    relPixelsX = xPixels - wcs%x%refPixels
    relPixelsY = yPixels - wcs%y%refPixels

    intermediateWorldX = wcs%x%matrixOnAxis  * relPixelsX&
                       + wcs%y%matrixOffAxis * relPixelsY
    intermediateWorldY = wcs%y%matrixOnAxis  * relPixelsY&
                       + wcs%x%matrixOffAxis * relPixelsX

    tanX = intermediateWorldX * intermediateWorldXToTanX
    tanY = intermediateWorldY * intermediateWorldYToTanY

  end subroutine pixelsToTanVector

  !---------------------------------------------------------------------
  subroutine pixelsToTanArray(xPixels, yPixels, wcs, tanX, tanY)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(single), intent(in)  :: xPixels(:,:),&
                                 yPixels(size(xPixels,1),size(xPixels,2))
    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) :: tanX(size(xPixels,1),size(xPixels,2)),&
                                 tanY(size(xPixels,1),size(xPixels,2))

    real(single), allocatable :: xPixelsVec(:), yPixelsVec(:)
    real(double), allocatable :: tanXVec(:), tanYVec(:)
    integer(int32) :: xSize, ySize, vSize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xPixels, 1)
    ySize = size(xPixels, 2)
    vSize = xSize * ySize
    allocate(xPixelsVec(vSize), yPixelsVec(vSize))
    allocate(tanXVec(vSize), tanYVec(vSize))

    xPixelsVec = reshape(xPixels, (/vSize/))
    yPixelsVec = reshape(yPixels, (/vSize/))

    call pixelsToTanVector(xPixelsVec, yPixelsVec, wcs, tanXVec, tanYVec)

    tanX = reshape(tanXVec, (/xSize, ySize/))
    tanY = reshape(tanYVec, (/xSize, ySize/))

  end subroutine pixelsToTanArray

  !---------------------------------------------------------------------
  subroutine tanToPixelsScalar(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX, tanY
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels, yPixels

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

    tanXVec(1) = tanX
    tanYVec(1) = tanY

    call tanToPixelsVector(wcs, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = xPixelsVec(1)
    yPixels = yPixelsVec(1)

  end subroutine tanToPixelsScalar

  !---------------------------------------------------------------------
  subroutine tanToPixelsVector(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is inverted in the present subroutine, after x_i is first converted
  ! from my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: tanX(:), tanY(size(tanX))
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels(size(tanX)), yPixels(size(tanX))

    real(double), allocatable ::&
      intermediateWorldX(:), intermediateWorldY(:)
    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY,&
      det, invMatrix(2,2)
    real(single), allocatable :: relPixelsX(:), relPixelsY(:)

    ! The above changes of sized arrays to allocatable was done in order
    ! to work around a compiler bug. If local arrays are declared to the
    ! size of an input array, this seems to work fine 99% of the time,
    ! but once in a while results in a seg fault at the instant the
    ! subroutine is called. On these occasions no executable statements
    ! of the subroutine are executed before the seg fault. Previously I
    ! have experienced this trouble just with user-defined structure
    ! arrays, now I get it with simple real kind arrays. IMS 2004/04/27.

    character(3) :: xProjTypeStr, yProjTypeStr
    logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa tanToPixelsVector'

    allocate(intermediateWorldX(size(tanX)), intermediateWorldY(size(tanX)))
    allocate(relPixelsX(size(tanX)), relPixelsY(size(tanX)))

!    call checkWcsType(wcs)
    xProjTypeStr = getProjectionType(wcs%x)
    yProjTypeStr = getProjectionType(wcs%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:tanToPixelsVector: Non-TAN &
&projections are not yet supported.")

if (testFlag) print *, '      bbb'

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select
if (testFlag) print *, '      ccc'

    select case(wcs%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
&than 'deg' are not yet supported.")
    end select
if (testFlag) print *, '      ddd'

    intermediateWorldX = tanX / intermediateWorldXToTanX
    intermediateWorldY = tanY / intermediateWorldYToTanY
if (testFlag) print *, '      eee'

    det = wcs%x%matrixOnAxis  * wcs%y%matrixOnAxis&
        - wcs%x%matrixOffAxis * wcs%y%matrixOffAxis
    if (det == 0.0)&
call error('singularWcsRotMatrix',"ssclib:pixelsToTanVector: Can't invert the WCS rotation matrix")
if (testFlag) print *, '      fff'

    invMatrix(1,1) =  wcs%y%matrixOnAxis  / det		! 1st index row,
    invMatrix(1,2) = -wcs%y%matrixOffAxis / det		! 2nd index col.
    invMatrix(2,1) = -wcs%x%matrixOffAxis / det
    invMatrix(2,2) =  wcs%x%matrixOnAxis  / det
if (testFlag) print *, '      ggg'

    relPixelsX = invMatrix(1,1) * intermediateWorldX&
               + invMatrix(1,2) * intermediateWorldY
    relPixelsY = invMatrix(2,1) * intermediateWorldX&
               + invMatrix(2,2) * intermediateWorldY
if (testFlag) print *, '      hhh'

    xPixels = relPixelsX + wcs%x%refPixels
    yPixels = relPixelsY + wcs%y%refPixels
if (testFlag) print *, '      zzz tanToPixelsVector'

  end subroutine tanToPixelsVector

  !---------------------------------------------------------------------
  subroutine tanToPixelsArray(wcs, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX(:,:),&
                                 tanY(size(tanX,1),size(tanX,2))
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: xPixels(size(tanX,1),size(tanX,2)),&
                                 yPixels(size(tanX,1),size(tanX,2))

    real(single), allocatable :: xPixelsVec(:), yPixelsVec(:)
    real(double), allocatable :: tanXVec(:), tanYVec(:)
    integer(int32) :: xSize, ySize, vSize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanX, 1)
    ySize = size(tanX, 2)
    vSize = xSize * ySize
    allocate(xPixelsVec(vSize), yPixelsVec(vSize))
    allocate(tanXVec(vSize), tanYVec(vSize))

    tanXVec = reshape(tanX, (/vSize/))
    tanYVec = reshape(tanY, (/vSize/))

    call tanToPixelsVector(wcs, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = reshape(xPixelsVec, (/xSize, ySize/))
    yPixels = reshape(yPixelsVec, (/xSize, ySize/))

  end subroutine tanToPixelsArray

  !---------------------------------------------------------------------
  subroutine tanRadToPixEllipseScalar(radiusTan, wcs, shortSemiAxisPixels&
  , longSemiAxisPixels, rotatedAngle)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: radiusTan
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: shortSemiAxisPixels,&
                                 longSemiAxisPixels,&
                                 rotatedAngle

    real(single) ::&
      shortSemiAxisPixelsVec(1), longSemiAxisPixelsVec(1), rotatedAngleVec(1)

    !-------------------------------------------------------------------
    ! Start

    call tanRadToPixEllipseVector((/radiusTan/), wcs, shortSemiAxisPixelsVec&
      , longSemiAxisPixelsVec, rotatedAngleVec)

    shortSemiAxisPixels = shortSemiAxisPixelsVec(1)
    longSemiAxisPixels  = longSemiAxisPixelsVec(1)
    rotatedAngle        = rotatedAngleVec(1)

  end subroutine tanRadToPixEllipseScalar

  !---------------------------------------------------------------------
  subroutine tanRadToPixEllipseVector(radiusTan, wcs, shortSemiAxisPixels&
  , longSemiAxisPixels, rotatedAngle)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: radiusTan(:)
    type(WcsT),   intent(in)  :: wcs
    real(single), intent(out) :: shortSemiAxisPixels(size(radiusTan)),&
                                 longSemiAxisPixels(size(radiusTan)),&
                                 rotatedAngle(size(radiusTan))

    character(3) :: xProjTypeStr, yProjTypeStr
    real(double) :: intermediateWorldXToTanX, xPixelSizeWorld, yPixelSizeWorld, rotAngleDble
!, intermediateWorldYToTanY
    real(double), allocatable :: radiusWorld(:)

    !-------------------------------------------------------------------
    ! Start

    xProjTypeStr = getProjectionType(wcs%x)
    yProjTypeStr = getProjectionType(wcs%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:tanRadToPixEllipseVector: Non-TAN &
&projections are not yet supported.")

    select case(wcs%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select

!    select case(wcs%y%worldUnit)
!    case('deg')
!      intermediateWorldYToTanY = PiOn180
!    case default
!call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
!&than 'deg' are not yet supported.")
!    end select

    allocate(radiusWorld(size(radiusTan)))
!, rotAngleDble(size(radiusTan)))
!    allocate(xPixelSizeWorld(size(radiusTan)), yPixelSizeWorld(size(radiusTan)))

    radiusWorld = radiusTan / intermediateWorldXToTanX
    call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngleDble)
    rotatedAngle = real(rotAngleDble, single)
    shortSemiAxisPixels = radiusWorld / xPixelSizeWorld
    longSemiAxisPixels  = radiusWorld / yPixelSizeWorld

!***** need to check this routine and also fix up the tan->world bit. I wrote it in a hurry and my head is a bit dull today.

  end subroutine tanRadToPixEllipseVector

  !---------------------------------------------------------------------
  logical(bool) function wcsAreEquivalent(wcs1, wcs2)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs1, wcs2

    !-------------------------------------------------------------------
    ! Start

    wcsAreEquivalent = .false. ! default

  if (wcs1%x%ctype          .ne. wcs2%x%ctype         ) return
  if (.not. approxEqual(wcs1%x%refPixels,      wcs2%x%refPixels)     ) return
  if (.not. approxEqual(wcs1%x%refWorld,       wcs2%x%refWorld)      ) return

  if (wcs1%y%ctype          .ne. wcs2%y%ctype         ) return
  if (.not. approxEqual(wcs1%y%refPixels,      wcs2%y%refPixels)     ) return
  if (.not. approxEqual(wcs1%y%refWorld,       wcs2%y%refWorld)      ) return

!*** fix up

    wcsAreEquivalent = .true.

  end function wcsAreEquivalent

  !---------------------------------------------------------------------
  subroutine convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    type(WcsT),   intent(inout) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%x%matrixOnAxis   =  xPixelSizeWorld * cos(rotAngle)
    wcs%x%matrixOffAxis  =  xPixelSizeWorld * sin(rotAngle)
    wcs%y%matrixOffAxis  = -yPixelSizeWorld * sin(rotAngle)
    wcs%y%matrixOnAxis   =  yPixelSizeWorld * cos(rotAngle)

  end subroutine convertOldToNew

  !---------------------------------------------------------------------
  subroutine convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    type(WcsT),   intent(in) :: wcs
    real(double), intent(out) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    real(double) :: rotAngleAlt, rotAngle2

    !-------------------------------------------------------------------
    ! Start:

    xPixelSizeWorld = sqrt(wcs%X%matrixOnAxis * wcs%X%matrixOnAxis&
      + wcs%X%matrixOffAxis * wcs%X%matrixOffAxis)

    yPixelSizeWorld = sqrt(wcs%Y%matrixOnAxis * wcs%Y%matrixOnAxis&
      + wcs%Y%matrixOffAxis * wcs%Y%matrixOffAxis)

    rotAngle    = atan2(-wcs%Y%matrixOffAxis, wcs%Y%matrixOnAxis)
    rotAngleAlt = atan2( wcs%X%matrixOffAxis, wcs%X%matrixOnAxis)

    if (abs(rotAngleAlt - rotAngle) > PiOn2) then ! one of the pixel
    ! dimensions should be negative. It is immaterial which, but I'll
    ! pick the x-axis dimension as recommended by C&G:
      xPixelSizeWorld = -xPixelSizeWorld

      ! Just as a check:

      rotAngle2 = atan2(-wcs%X%matrixOffAxis, -wcs%X%matrixOnAxis)
      if (abs(rotAngle2 - rotAngle) > PiOn2)&
call error('wcs_aux_bug0',"ssclib:convertNewToOld: This should not be able &
&to happen. Please contact the code developer.")
    end if

  end subroutine convertNewToOld

  !---------------------------------------------------------------------
  function wcsPixelAreaWorld(wcs) result(area)
  !---------------------------------------------------------------------
  ! Calculates the area in world coordinates of an image pixel.

    implicit none

    type(WcsT), intent(in) :: wcs
    real(double)           :: area

    real(single) :: squareXPixels(2,2), squareYPixels(2,2)
    real(double) ::&
      relPixelsX(2,2), relPixelsY(2,2),&
      parlogramXWorld(2,2), parlogramYWorld(2,2),&
      vectorA(3), vectorB(3), crossProd(3)

    !-------------------------------------------------------------------
    ! Start

    squareXPixels(:,2) = (/0.0, 1.0/)
    squareXPixels(:,1) = (/0.0, 1.0/)

    squareYPixels(:,2) = (/1.0, 1.0/)
    squareYPixels(:,1) = (/0.0, 0.0/)

    relPixelsX = squareXPixels - wcs%x%refPixels
    relPixelsY = squareYPixels - wcs%y%refPixels

    parlogramXWorld = wcs%x%matrixOnAxis  * relPixelsX&
                    + wcs%y%matrixOffAxis * relPixelsY
    parlogramYWorld = wcs%y%matrixOnAxis  * relPixelsY&
                    + wcs%x%matrixOffAxis * relPixelsX

    ! The transformed square has now (in general) been sheared into a
    ! parallelogram, then rotated and offset. We'll now load the sides
    ! of the transformed square into two vectors. Note that we need a z
    ! component in order to be able to do a cross product; but of course
    ! since all the above operations take place in the plane, z components
    ! are set to zero.

    vectorA = (/parlogramXWorld(2,1) - parlogramXWorld(1,1)&
              , parlogramYWorld(2,1) - parlogramYWorld(1,1), 0.0d0/)

    vectorB = (/parlogramXWorld(1,2) - parlogramXWorld(1,1)&
              , parlogramYWorld(1,2) - parlogramYWorld(1,1), 0.0d0/)

    ! Area of a parallelogram defined by two vectors is just the magnitude
    ! of their cross product:

    crossProd = vectorCross(vectorA, vectorB)
    area = sqrt(sum(crossProd * crossProd))

  end function wcsPixelAreaWorld

  !---------------------------------------------------------------------
  subroutine checkImageEquivalence(imageSetNames, equivalent, xSize, ySize, wcs)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: imageSetNames(:)
    logical,        intent(out) :: equivalent
    integer(int16), intent(out) :: xSize, ySize
    type(WcsT),     intent(out) :: wcs

    integer(int16) :: numImages, imageNum, tempXSize, tempYSize

    !-------------------------------------------------------------------
    ! Start

    equivalent = .true. ! default

    xSize = 0 ! Don't like to leave
    ySize = 0 ! them unset.

    call initializeWcsT(wcs)

    numImages = size(imageSetNames)
  if (numImages <= 0) return

call readWcs(imageSetNames(1), wcs)
!*** should also read the wcs from each of the other arrays and test them for equivalence. But there are still problems with the wcs equiv test.

    call getArraySize(imageSetNames(1), xSize, ySize) ! in ssclib/dal_aux
    do imageNum = 2, numImages
      call getArraySize(imageSetNames(imageNum), tempXSize, tempYSize)
      if (tempXSize .ne. xSize .or. tempYSize .ne. ySize) then
        equivalent = .false.
  return
      end if
    end do

  end subroutine checkImageEquivalence

  !---------------------------------------------------------------------
  function getProjectionType(wcsAxis) result(projectionTypeStr)
  !---------------------------------------------------------------------

    implicit none

    type(WcsAxisT), intent(in) :: wcsAxis
    character(3)               :: projectionTypeStr

    !-------------------------------------------------------------------
    ! Start

    projectionTypeStr = wcsAxis%ctype(6:8)

  end function getProjectionType

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module wcs_aux
